# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.api`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Base RightScale API Access Object.

This package provides access to the RightScale API via Tornado-style
`@gen.coroutine` wrapped methods. These methods are, however, just wrappers
for threads that are being fired off in the background to make the API
calls.

**Async vs Threads**

In the future, this will get re-factored to use a native Tornado
AsyncHTTPClient object. The methods themselves will stay the same, but the
underlying private methods will change.

The methods in this object are specifically designed to support common
operations that the RightScale Actor objects need to do. Operations like
'find server array', 'launch server array', etc. This is not meant as a pure
one-to-one mapping of the RightScale API, but rather a mapping of conceptual
operations that the Actors need.

**Method Design Note**

RightScale mixes and matches their API calls... some of them you pass in a
major method and then supply a resource ID to act on. Others you pass in the
resource_id and get back a list of methods that you can execute.

For consistency in our programming model, this class relies o you passing in
rightscale.Resource objects everywhere, and it does the resource->ID
translation.
"""

from datetime import datetime
from os import path
from six import string_types
import functools
import logging

from retrying import retry as sync_retry
from rightscale import util as rightscale_util
from tornado import concurrent
from tornado import gen
from tornado import ioloop
import requests
import rightscale
import simplejson

from kingpin import utils
from kingpin.actors.rightscale import settings

log = logging.getLogger(__name__)

# Suppress InsecurePlatformWarning
requests.packages.urllib3.disable_warnings()

__author__ = 'Matt Wise <matt@nextdoor.com>'


DEFAULT_ENDPOINT = 'https://my.rightscale.com'

# This executor is used by the tornado.concurrent.run_on_executor()
# decorator. We would like this to be a class variable so its shared
# across RightScale objects, but we see testing IO errors when we
# do this.
EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)


class RightScaleError(Exception):

    """Raised when the RightScale API returns a specific error"""


def rightscale_error_logger(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except requests.exceptions.HTTPError as e:
            log.error('Error in RightScale API Call: %s(%s, %s): %s'
                      % (func.__name__, args, kwargs, e))

            if hasattr(e, 'response') and hasattr(e.response, 'text'):
                raise RightScaleError('RightScale Error: %s' % e.response.text)

            raise
    return wrapper


class ServerArrayException(Exception):

    """Raised when an operation on or looking for a ServerArray fails"""


class RightScale(object):

    # Get references to existing objects that are used by the
    # tornado.concurrent.run_on_executor() decorator.
    ioloop = ioloop.IOLoop.current()
    executor = EXECUTOR

    def __init__(self, token, endpoint=DEFAULT_ENDPOINT):
        """Initializes the RightScaleOperator Object for a RightScale Account.

        Args:
            token: A RightScale RefreshToken
            api: API URL Endpoint
        """
        self._token = token
        self._endpoint = endpoint
        self._client = rightscale.RightScale(refresh_token=self._token,
                                             api_endpoint=self._endpoint)

        # Quiet down the urllib requests library, its noisy even in
        # INFO mode and muddies up the logs.
        r_log = logging.getLogger('requests.packages.urllib3.connectionpool')
        r_log.setLevel(logging.WARNING)

        log.debug('%s initialized (token=<hidden>, endpoint=%s)' %
                  (self.__class__.__name__, endpoint))

    def get_res_id(self, resource):
        """Returns the Resource ID of a given RightScale Resource object.

        Args:
            rightscale.Resource object

        Returns:
            Integer of Resource ID
        """
        return int(path.split(resource.self.path)[-1])

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def find_server_arrays(self, name, exact=True):
        """Search for a list of ServerArray by name and return the resources.

        Args:
            name: RightScale ServerArray Name
            exact: Return a single exact match, or multiple matching resources.

        Returns:
            <rightscale.Resource object(s)>
        """
        log.debug('Searching for ServerArrays matching: %s (exact match: %s)' %
                  (name, exact))

        found_arrays = rightscale_util.find_by_name(
            self._client.server_arrays, name, exact=exact)

        if not found_arrays:
            log.debug('ServerArray matching "%s" not found' % name)
            return

        if isinstance(found_arrays, list):
            names = [s.soul['name'] for s in found_arrays]
        else:
            names = [found_arrays.soul['name']]

        log.debug('Got ServerArray(s): %s' % ', '.join(names))

        return found_arrays

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def show(self, resource):
        """Async wrapping of <resource>.show() with retry wrapper.

        Args:
            resource: rightscale.Resource object

        Returns:
            <rightscale.Resource object>.show()
        """
        return resource.show()

    @concurrent.run_on_executor
    @rightscale_error_logger
    @utils.exception_logger
    def find_cookbook(self, name):
        """Search for a Cookbook by-name and return the resource.

        Args:
            name: Cookbook Name

        Return:
            rightscale.Resource object
        """
        cookbook = name.split('::')[0]

        log.debug('Searching for Cookbooks matching: %s' % name)
        found_cookbooks = self._client.cookbooks.index(
            params={'filter[]': ['name==%s' % cookbook],
                    'view': 'extended'})
        found_recipes = [r for r in found_cookbooks
                         if r.soul['metadata']['recipes'].get(name)]

        if not found_recipes:
            log.debug('Recipe matching "%s" could not be found.' % name)
            log.debug('Found cookbooks %s' % found_cookbooks)
            return

        recipe = found_recipes[0]

        log.debug('Found recipe: %s' % recipe)

        return recipe

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def find_right_script(self, name):
        """Search for a RightScript by-name and return the resource.

        Args:
            name: RightScale RightScript Name

        Return:
            rightscale.Resource object
        """
        log.debug('Searching for RightScript matching: %s' % name)
        found_script = rightscale_util.find_by_name(
            self._client.right_scripts, name, exact=True)

        if not found_script:
            log.debug('RightScript matching "%s" could not be found.' % name)
            return

        log.debug('Got RightScript: %s' % found_script)

        return found_script

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def find_by_name_and_keys(self, collection, exact=True, **kwargs):
        """Search for a RightScale resource by name, and optional keys.

        This code is blatently stolen from rightscale.util.find_by_name and
        just re-worked so that we can search with the subject_href.
        RightScale deliberately clones AlertSpecs all of the place. For our
        purposes, searching with the subject_href becomes a requirement to
        avoid complex scenarios where we may return the wrong AlertSpec.

        Args:
            collection: RightScale.<xxx> resource object
            exact: If True, returns the first match. If False, returns a list
                of all returned resources.
            **kwargs: Any additional keys-and-values to use in the search.

        Returns:
            One RightScale Resource Object or a List of objects.
        """
        filter_keys = []
        for key, val in list(kwargs.items()):
            filter_keys.append('%s==%s' % (key, val))
        params = {'filter[]': sorted(filter_keys)}

        found = collection.index(params=params)
        if not exact and len(found) > 0:
            return found

        if len(found) < 1:
            return []

        if len(found) == 1:
            return found[0]

        return found

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def destroy_resource(self, res):
        """Destroy an RightScale resource.

        Args:
            res: Resource object to destroy
        """
        return res.self.destroy()

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def create_resource(self, res, params):
        """Create an RightScale resource.

        Args:
            res: Resource object to destroy
            params: Dict of RightScale parameters to pass in

        Returns:
            The Rightscale Resource itself
        """
        return res.create(params=params)

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def commit_resource(self, res, res_type, message=None, params=None):
        """Commit a RightScale resource

        Args:
            res: Resource object to commit
            res_type: The RightScale resource object _type_
            message: The message to use when committing
            params: A dictionary of pre-filled out parameters for the commit
                    call.

        Returns:
            The Rightscale Resource itself
        """
        res_id = self.get_res_id(res)
        if not params:
            params = {'commit_message': message}
        return res_type.commit(res_id=res_id, params=params)

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def add_resource_tags(self, res, tags):
        """Tags a RightScale resource

        Args:
            res: Resource object to commit
            tag: The tag(s) to add to the resource

        Returns:
            The Rightscale Resource itself
        """
        params = [('resource_hrefs[]', res.href)]
        for tag in tags:
            params.append(('tags[]', tag))
        return self._client.tags.multi_add(params=params)

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def delete_resource_tags(self, res, tags):
        """Deletes tags from a RightScale resource

        Args:
            res: Resource object to commit
            tag: The tag(s) to delete from the resource

        Returns:
            The Rightscale Resource itself
        """
        params = [('resource_hrefs[]', res.href)]
        for tag in tags:
            params.append(('tags[]', tag))
        return self._client.tags.multi_delete(params=params)

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def get_resource_tags(self, res):
        """Returns a list of tags associated with a RightScale resource.

        Args:
            res: Resource object to search for

        Returns:
            [List, of, tags]
        """
        params = [('resource_hrefs[]', res.href)]
        raw = self._client.tags.by_resource(params=params)[0]
        tags = [tag['name'] for tag in raw.soul['tags']]
        return tags

    @concurrent.run_on_executor
    @rightscale_error_logger
    @utils.exception_logger
    def clone_server_array(self, array):
        """Clone a Server Array.

        Clones an existing Server Array into a new array. Requires the
        source template array ID number. Returns the newly cloned array.

        Args:
            array: Source ServerArray Resource Object

        Return:
            <rightscale.Resource object>
        """
        log.debug('Cloning ServerArray %s' % array.soul['name'])
        source_id = self.get_res_id(array)
        new_array = self._client.server_arrays.clone(res_id=source_id)
        log.debug('New ServerArray %s created!' % new_array.soul['name'])
        return new_array

    @concurrent.run_on_executor
    @rightscale_error_logger
    @utils.exception_logger
    def destroy_server_array(self, array):
        """Destroys a Server Array.

        Makes this API Call:

            http://reference.rightscale.com/api1.5/resources/
            ResourceServerArrays.html#destroy

        Args:
            array: ServerArray Resource Object
        """
        log.debug('Destroying ServerArray %s' % array.soul['name'])
        array_id = self.get_res_id(array)
        self._client.server_arrays.destroy(res_id=array_id)
        log.debug('Array Destroyed')

    @concurrent.run_on_executor
    @rightscale_error_logger
    @utils.exception_logger
    def update(self, resource, params, sub_resource='self'):
        """Updates a RightScale resource with the supplied parameters.

        Valid parameters can be found at the following URL:

            http://reference.rightscale.com/api1.5/resources/
            ResourceServerArrays.html#update
            http://reference.rightscale.com/api1.5/resources/
            ResourceInstances.html#update

        Args:
            resource: rightscale.Resource object to update.
            params: The parameters to update. eg:
                { 'server_array[name]': 'new name' }
            sub_resource: Optionally, rather than calling resource.self.update,
                you can pass in a sub_resource and it will call
                resource.<sub_resource>.update. Useful for updating things like
                RightScripts that have special URL endpoints for updates.

        Returns:
            <updated rightscale array object>
        """

        log.debug('Resource: %s' % resource)
        r = getattr(resource, sub_resource)

        if isinstance(params, string_types):
            r.update(data=params)
        else:
            r.update(params=params)

        updated_resource = resource.self.show()
        return updated_resource

    @concurrent.run_on_executor
    @rightscale_error_logger
    @utils.exception_logger
    def get_server_array_inputs(self, array):
        """Looks up ServerArray 'Next Instance' inputs.

        Valid parameters can be found at the following URL:

            http://reference.rightscale.com/api1.5/resources/
            ResourceInputs.html#index

        Args: rightscale.Resource array object.
        Returns:
            List of rightscale.Resource input objects.
        """
        instance = array.next_instance.show()
        all_inputs = instance.inputs.index()

        return all_inputs

    @concurrent.run_on_executor
    @rightscale_error_logger
    @utils.exception_logger
    def update_server_array_inputs(self, array, inputs):
        """Updates a ServerArray 'Next Instance' with the supplied inputs.

        Valid parameters can be found at the following URL:

            http://reference.rightscale.com/api1.5/resources/
            ResourceInputs.html#multi_update

        Note: Its impossible to tell whether the update has succeeded because
        the RightScale API always returns a '204 No Content' message on the
        multi_update() call. Therefore, we simply execute the command return.

        Args:
            array: rightscale.Resource object to update.
            inputs: The parameters to update. eg:
                { 'inputs[ELB_NAME]': 'text:foobar' }
        """

        log.debug('Patching ServerArray (%s) with new inputs: %s' %
                  (array.soul['name'], inputs))

        next_inst = array.next_instance.show()
        next_inst.inputs.multi_update(params=inputs)

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def launch_server_array(self, array, count=1):
        """Launches an instance of a ServerArray..

        Makes this API Call:

            http://reference.rightscale.com/api1.5/resources/
            ResourceServerArrays.html#launch

        Note: Repeated simultaneous calls to this method on the same array will
        return 422 errors from RightScale. It is advised that you make this
        call synchronously on a particular array as many times as you need.
        This method is wrapped in a retry block though to help handle these
        errors anyways.

        Args:
            array: ServerArray Resource Object
            count: Instances to launch (default: 1)

        Returns:
            rightscale.Resource of the newly launched instance>
        """
        if not count or count < 1:
            return

        # The RightScale API supports sending in a 'count' to launch many
        # servers at once. This is only functional though if you submit a count
        # of > 1. Otherwise, it fails.
        params = None
        if count > 1:
            params = {'count': count}

        log.debug('Launching a new instance of ServerArray %s' %
                  array.soul['name'])
        array_id = self.get_res_id(array)
        return self._client.server_arrays.launch(
            res_id=array_id, params=params)

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def get_server_array_current_instances(
            self, array, filters=['state<>terminated']):
        """Returns a list of ServerArray current running instances.

        Makes this API Call:

            http://reference.rightscale.com/api1.5/resources/
            ResourceServerArrays.html#current_instances

        Valid Filters:

            http://reference.rightscale.com/api1.5/resources/
            ResourceInstances.html#index_filters

        Args:
            array: rightscale.Resource object to count
            filters: List of filters to use to find instances.

        Returns:
            [<list of rightscale.Resource objects>]
        """
        log.debug('Searching for current instances of ServerArray (%s)' %
                  array.soul['name'])
        params = {'filter[]': filters}
        return array.current_instances.index(params=params)

    @concurrent.run_on_executor
    @rightscale_error_logger
    @utils.exception_logger
    def terminate_server_array_instances(self, array):
        """Executes a terminate on all of the current running instances.

        Makes this API Call:

            http://reference.rightscale.com/api1.5/resources/
            ResourceServerArrays.html#multi_terminate

        Returns as soon as RightScale claims that the operation is completed --
        but this only means that the servers have been 'told' to shut down, not
        that they are actually terminated yet.

        Args:
            array: ServerArray Resource Object

        Return:
            <task object for termination request>
        """
        log.debug('Terminating all instances of ServerArray (%s)' %
                  array.soul['name'])
        array_id = self.get_res_id(array)
        try:
            task = self._client.server_arrays.multi_terminate(res_id=array_id)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 422:
                # There are no instances to terminate.
                return

        return task

    @gen.coroutine
    def wait_for_task(self,
                      task,
                      task_name=None,
                      sleep=5,
                      loc_log=log,
                      instance=None):
        """Monitors a RightScale task for completion.

        RightScale tasks are provided as URLs that we can query for the
        run-status of the task. This method repeatedly queries a task for
        completion (every 5 seconds), and returns when the task has finished.

        TODO: Add a task-timeout option.

        Note: This is a completely retryable operation in the event that an
        intermittent network connection causes any kind of a connection
        failure.

        Args:
            task: RightScale Task resource object.
            task_name: Human-readable name of the task to be executed.
            sleep: Integer of seconds to wait before the first status check.
            loc_log: logging.getLogger() object to be used to log task status.
                    This is useful when this API call is called from a Kingpin
                    actor, and you want to use the actor's specific logger.
                    If nothing is passed - local `log` object is used.
            instance: RightScale instance object on which the task is executed.

        Returns:
            bool: success status
        """

        if not task:
            # If there is no task to wait on - don't wait!
            raise gen.Return(True)

        timeout_id = None
        if task_name:
            timeout_id = utils.create_repeating_log(
                loc_log.info, 'Still waiting on %s' % task_name, seconds=sleep)

        # Tracking when the tasks start so we can search by date later
        # RightScale expects the time to be a string in UTC
        now = datetime.utcnow()
        tasks_start = now.strftime('%Y/%m/%d %H:%M:%S +0000')

        while True:
            # Get the task status
            output = yield self._get_task_info(task)
            summary = output.soul['summary'].lower()
            stamp = datetime.now()

            if 'success' in summary or 'completed' in summary:
                status = True
                break

            if 'failed' in summary:
                status = False
                break

            loc_log.debug('Task (%s) status: %s (updated at: %s)' %
                          (output.path, output.soul['summary'], stamp))

            yield utils.tornado_sleep(min(sleep, 5))

        loc_log.debug('Task (%s) status: %s (updated at: %s)' %
                      (output.path, output.soul['summary'], stamp))

        if timeout_id:
            utils.clear_repeating_log(timeout_id)

        if status is True:
            raise gen.Return(True)

        if not instance:
            raise gen.Return(status)

        # If something failed we want to find out why -- get audit logs

        # Contact RightScale for audit logs of this instance.
        now = datetime.utcnow()
        tasks_finish = now.strftime('%Y/%m/%d %H:%M:%S +0000')

        loc_log.error('Task failed. Instance: "%s".' % instance.soul['name'])

        audit_logs = yield self.get_audit_logs(
            instance=instance,
            start=tasks_start,
            end=tasks_finish,
            match='failed')

        # Print every audit log that was obtained (may be 0)
        [loc_log.error(l) for l in audit_logs]

        if not audit_logs:
            loc_log.error('No audit logs for %s' % instance)

        loc_log.debug('Task finished, return value: %s, summary: %s' %
                      (status, summary))

        raise gen.Return(status)

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def _get_task_info(self, task):
        """Fetch data for a particular RightScale task.

        This is a blocking, non-tornado operation. It's separated into its own
        function to be run on a separate thread.
        """
        return task.self.show()

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    @utils.exception_logger
    def get_audit_logs(self, instance, start, end, match=None):
        """Fetch a set of audit logs belonging to an instance.

        http://reference.rightscale.com/api1.5/resources/
        ResourceAuditEntries.html

        Args:
            instance: RightScale instance object.
            start: String as expected by start_date of the API
                   e.g., 2011/06/25 00:00:00 +0000.
            end: String as expected by end_date of the API.
            match: optional string to match the summary of the audit entry.
                   Only audit entries with this string will be returned.

        Returns:
            list of audit entries between the start and end date that match
            a substring in the summary. May return an empty list.

        """

        href = instance.links['self']
        all_entries = self._client.audit_entries.index(params={
            'filter[]': ['auditee_href==%s' % href],
            'limit': 10,
            'start_date': start,
            'end_date': end
        })

        log.debug('Found %s audit logs.' % len(all_entries))

        logs = []
        for entry in all_entries:
            summary = entry.soul['summary']
            if match and match not in summary:
                log.debug('Skipping details for "%s"' % summary)
                continue
            log.debug('Fetching details for "%s"' % summary)

            # grabbing raw output because RightScale doesn't reply via JSON
            # when accessing details of a log.
            detail_res = self._client.client.get(entry.detail.path)
            details = detail_res.raw_response.text

            logs.append(details)

        return logs

    @gen.coroutine
    def run_executable_on_instances(self, name, inputs, instances):
        """Execute a script on a set of RightScale Instances.

        This method bypasses the python-rightscale native properties and
        callable methods because they are broken with regards to running
        individual API calls against instances. See this bug:

            https://github.com/brantai/python-rightscale/issues/6

        Instead, we take in a list of rightscale.Resource objects that point to
        instances. For each instance we iterate over and directly call the
        <instance_path>/run_executable URL. This is done below in the
        make_generic_request() method for us.

        Note, the inputs dictionary should look like this:
            { '' }

        Args:
            name: Recipe or RightScript String Name
            inputs: Dict of Key/Value Input Pairs
            instances: A list of rightscale.Resource instances objects.

        Returns:
            list of tuples - (instance, <rightscale.Resource task object>)
        """
        # Create a new copy of the inputs that were passed in so that we can
        # modify them correctly and safely.
        params = dict(inputs)

        # Determine whether we're looking for a recipe or a rightscript. If its
        # the latter, we have to go and find its href identifier first.
        if '::' in name:
            script_type = 'Recipe'
            params['recipe_name'] = name
        else:
            script_type = 'RightScript'
            script = yield self.find_right_script(name)

            if not script:
                raise ServerArrayException('RightScript Not Found')

            params['right_script_href'] = script.href

        log.debug('Executing %s with params: %s' % (script_type, params))

        # Walk through the list of instances and fire off the execution on each
        # instance. For each execution, we will store a reference to the
        # instane itself, and the task thats executing. Note, as soon as we
        # call the make_generic_request() method, a thread is fired off and
        # begins acting on that request. Outside of this loop (below), we will
        # iterate over the responses to these requests.
        task_pairs = []
        for i in instances:
            log.debug('Executing %s on %s' % (name, i.soul['name']))
            url = '%s/run_executable' % i.links['self']
            req = self.make_generic_request(url, post=params)
            task_pairs.append((i, req))

        # At this point, all of our tasks are executing in the background. We
        # can now yield on each task *individually* in order to get the "result
        # object" back. This looks synchronous, but remember that the real API
        # calls are actually happening in the background simultaneously.
        yielded_tasks = []
        exceptions_caught = []
        for (i, task) in task_pairs:
            try:
                result = yield task
                yielded_tasks.append((i, result))
            except (requests.exceptions.HTTPError, RightScaleError) as e:
                msg = ('Failed to queue execution on %s: %s' %
                       (i.soul['name'], e))
                exceptions_caught.append(msg)

        # Rather than a single try/except and raising a group of exceptions,
        # Tornado's 'multi_future' method raises the first exception in a list
        # of tasks. This behavior is described in a bug, and we are working
        # with the Tornado team to try to come up with a reasonable solution.
        # Until then, we do this hackery to create a single exception from
        # many, and then raise that exception:
        #
        # https://github.com/tornadoweb/tornado/issues/1378
        if exceptions_caught:
            exc_string = ', '.join(exceptions_caught)
            exc_length = len(exceptions_caught)
            raise ServerArrayException('%s failures: %s' %
                                       (exc_length, exc_string))

        raise gen.Return(yielded_tasks)

    @concurrent.run_on_executor
    @sync_retry(**settings.RETRYING_SETTINGS)
    @rightscale_error_logger
    def make_generic_request(self, url, post=None):
        """Make a generic API call and return a Resource Object.

        This method is a bit hacky. It manually executes a REST call against
        the RightScale API and then attempts to build a custom
        rightscale.Resource object based on those return results. This allows
        us to support API calls that the current python-rightscale library does
        not currently support (like running an executable on an instance of an
        array).

        Args:
            url: String of the URL to call
            post: Optional POST Body Data

        Returns:
            <rightscale.Resource objects>
        """
        # Make the initial web call
        log.debug('Making generic API call: %s (%s)' % (url, post))

        # Here we're reaching into the rightscale client library and getting
        # access directly to its requests client object.
        if post is not None:
            response = self._client.client.post(url, data=post)
        else:
            response = self._client.client.get(url)

        # Now, if a location tag was returned to us, follow it and get the
        # newly returned response data
        loc = response.headers.get('location', None)
        if loc:
            response = self._client.client.get(loc)
            url = loc

        # Try to parse the JSON body. If no body was returned, this fails and
        # thats OK sometimes.
        try:
            soul = response.json()
        except simplejson.scanner.JSONDecodeError:
            log.debug('No JSON found. Returning the raw text')
            return response.raw_response.text

        # Now dig deep into the python rightscale library itself and create our
        # own Resource object by hand.
        resource = rightscale.rightscale.Resource(
            path=url,
            response=response,
            soul=soul,
            client=self._client.client)

        return resource



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.s3`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import json
import logging

from botocore.exceptions import ClientError, ParamValidationError
from tornado import concurrent
from tornado import gen
from inflection import camelize
import jsonpickle

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.utils import dry
from kingpin.actors.aws import base
from kingpin.constants import SchemaCompareBase
from kingpin.constants import REQUIRED
from kingpin.constants import STATE

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com'


# This executor is used by the tornado.concurrent.run_on_executor()
# decorator. We would like this to be a class variable so its shared
# across RightScale objects, but we see testing IO errors when we
# do this.
EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)


class InvalidBucketConfig(exceptions.RecoverableActorFailure):

    """Raised whenever an invalid option is passed to a Bucket"""


class PublicAccessBlockConfig(SchemaCompareBase):

    """Provides JSON-Schema based validation of the supplied Public Access
    Block Configuration..

    The S3 PublicAccessBlockConfiguration should look like this:

    .. code-block:: json

        { "block_public_acls": true,
          "ignore_public_acls": true,
          "block_public_policy": true,
          "restrict_public_buckets": true }

    If you supply an empty dict, then we will explicitly remove the Public
    Access Block Configuration.

    """

    ACCESS_BLOCK_SCHEMA = {
        'type': ['object'],
        'required': [
            'block_public_acls',
            'ignore_public_acls',
            'block_public_policy',
            'restrict_public_buckets'
        ],
        'additionalProperties': False,
        'properties': {
            'block_public_acls': {'type': 'boolean'},
            'ignore_public_acls': {'type': 'boolean'},
            'block_public_policy': {'type': 'boolean'},
            'restrict_public_buckets': {'type': 'boolean'},
        }
    }

    SCHEMA = {
        'definitions': {
            'public_access_block_config': ACCESS_BLOCK_SCHEMA,
        },

        'anyOf': [
            {'$ref': '#/definitions/public_access_block_config'},
            {'type': 'null'},
            {'type': 'object', 'additionalProperties': False}
        ]
    }

    valid = (
        '{ "block_public_acls": true, "ignore_public_acls": false, '
        '"block_public_policy": true, "restrict_public_buckets": false }')


class LoggingConfig(SchemaCompareBase):

    """Provides JSON-Schema based validation of the supplied logging config.

    The S3 LoggingConfig format should look like this:

    .. code-block:: json

        { "target": "s3_bucket_name_here",
          "prefix": "an_optional_prefix_here" }

    If you supply an empty `target`, then we will explicitly remove the logging
    configuration from the bucket. Example:

    .. code-block:: json

        { "target": "" }

    """

    SCHEMA = {
        'type': ['object', 'null'],
        'required': ['target'],
        'additionalProperties': False,
        'properties': {
            'target': {'type': 'string'},
            'prefix': {'type': 'string'}
        }
    }

    valid = '{ "target": "<bucket name>", "prefix": "<logging prefix>" }'


class LifecycleConfig(SchemaCompareBase):

    """Provides JSON-Schema based validation of the supplied Lifecycle config.

    The S3 Lifecycle system allows for many unique configurations. Each
    configuration object defined in this schema will be turned into a
    :py:class:`boto.s3.lifecycle.Rule` object. All of the rules together will
    be turned into a :py:class:`boto.s3.lifecycle.Lifecycle` object.

    .. code-block:: json

        [
          {
            "id": "unique_rule_identifier",
            "status": "Enabled",
            "filter": {
              "prefix": "/some_path"
            },
            "transitions": [
              {
                "days": 90,
                "date": "2016-05-19T20:04:17+00:00",
                "storage_class": "GLACIER",
              }
            ],
            "noncurrent_version_transitions": [
              {
                "noncurrent_days": 90,
                "storage_class": "GLACIER",
              }
            ],
            "expiration": {
              "days": 365,
            },
            "noncurrent_version_expiration": {
              "noncurrent_days": 365,
            }
          }
        ]
    """

    SCHEMA = {
        'definitions': {
            'tag': {
                'type': 'object',
                'required': ['key', 'value'],
                'additionalProperties': False,
                'properties': {
                    'key': {
                        'type': 'string',
                    },
                    'value': {
                        'type': 'string',
                    },
                }
            },
            'transition': {
                'type': 'object',
                'required': ['storage_class'],
                'additionalProperties': False,
                'properties': {
                    'days': {
                        'type': ['string', 'integer'],
                        'pattern': '^[0-9]+$',
                    },
                    'date': {
                        'type': 'string',
                        'format': 'date-time'
                    },
                    'storage_class': {
                        'type': 'string',
                        'enum': ['GLACIER', 'STANDARD_IA']
                    }
                }
            },
            'noncurrent_version_transition': {
                'type': 'object',
                'required': ['storage_class'],
                'additionalProperties': False,
                'properties': {
                    'noncurrent_days': {
                        'type': ['string', 'integer'],
                        'pattern': '^[0-9]+$',
                    },
                    'storage_class': {
                        'type': 'string',
                        'enum': ['GLACIER', 'STANDARD_IA']
                    }
                }
            }
        },
        # The outer wrapper must be a list of properly formatted objects,
        # or Null if we are not going to manage this configuration at all.
        'type': ['array', 'null'],
        'uniqueItems': True,
        'items': {
            'type': 'object',
            'required': ['id', 'status'],
            'oneOf': [
                {'required': ['filter']},
                {'required': ['prefix']}
            ],
            'anyOf': [
                {
                    'oneOf': [
                        {'required': ['transition']},
                        {'required': ['transitions']}
                    ]
                },
                {
                    'oneOf': [
                        {'required': ['noncurrent_version_transition']},
                        {'required': ['noncurrent_version_transitions']}
                    ]
                },
                {'required': ['expiration']},
                {'required': ['noncurrent_version_expiration']},
                {'required': ['abort_incomplete_multipart_upload']}
            ],
            'additionalProperties': False,
            'properties': {
                # Basic Properties
                'id': {
                    'type': 'string',
                    'minLength': 1,
                    'maxLength': 255,
                },
                'status': {
                    'type': 'string',
                    'enum': ['Enabled', 'Disabled'],
                },

                # Filtering Properties
                #
                # prefix is deprecated in the AWS s3 API. Please use filter
                # instead.
                'filter': {
                    'type': 'object',
                    'minProperties': 1,
                    'maxProperties': 1,
                    'additionalProperties': False,
                    'properties': {
                        'prefix': {
                            'type': 'string',
                        },
                        'tag': {
                            '$ref': '#/definitions/tag'
                        },
                        'and': {
                            'type': 'object',
                            'minProperties': 1,
                            'maxProperties': 2,
                            'additionalProperties': False,
                            'properties': {
                                'prefix': {
                                    'type': 'string',
                                },
                                'tag': {
                                    '$ref': '#/definitions/tag'
                                },
                            }
                        }
                    }
                },
                'prefix': {
                    'type': 'string',
                },

                # Action Properties
                #
                # transition is deprecated in the AWS s3 API. Please use
                # transitions instead.
                'transitions': {
                    'type': 'array',
                    'itmes': {
                        '$ref': '#/definitions/transition'
                    }
                },
                'transition': {
                    '$ref': '#/definitions/transition'
                },
                # noncurrent_version_transition is deprecated in the AWS s3
                # API. Please use noncurrent_version_transitions instead.
                'noncurrent_version_transitions': {
                    'type': 'array',
                    'itmes': {
                        '$ref': '#/definitions/noncurrent_version_transition'
                    }
                },
                'noncurrent_version_transition': {
                    '$ref': '#/definitions/noncurrent_version_transition'
                },
                # Note for expireation, we allow the actor to just accept a
                # number of days instead of an object and we create the
                # correct json with days in the init. Hence the object type of
                # str/int/obj here.
                'expiration': {
                    'type': ['string', 'integer', 'object'],
                    'pattern': '^[0-9]+$',
                    'additionalProperties': False,
                    'properties': {
                        'days': {
                            'type': ['string', 'integer'],
                            'pattern': '^[0-9]+$',
                        },
                        'date': {
                            'type': 'string',
                            'format': 'date-time',
                        },
                        'expired_object_delete_marker': {
                            'type': 'boolean',
                        }
                    }
                },
                'noncurrent_version_expiration': {
                    'type': 'object',
                    'required': ['noncurrent_days'],
                    'additionalProperties': False,
                    'properties': {
                        'noncurrent_days': {
                            'type': ['string', 'integer'],
                            'pattern': '^[0-9]+$',
                        },
                    }
                },
                'abort_incomplete_multipart_upload': {
                    'type': 'object',
                    'required': ['days_after_initiation'],
                    'additionalProperties': False,
                    'properties': {
                        'days_after_initiation': {
                            'type': ['string', 'integer'],
                            'pattern': '^[0-9]+$',
                        },
                    }
                },
            }
        }
    }


class TaggingConfig(SchemaCompareBase):

    """Provides JSON-Schema based validation of the supplied tagging config.

    The S3 TaggingConfig format should look like this:

    .. code-block:: json

        [ { "key": "my_key", "value": "some_value" } ]

    """

    SCHEMA = {
        'type': ['array', 'null'],
        'uniqueItems': True,
        'items': {
            'type': 'object',
            'required': ['key', 'value'],
            'additionalProperties': False,
            'properties': {
                'key': {
                    'type': 'string',
                },
                'value': {
                    'type': 'string',
                }
            }
        }
    }

    valid = '[ { "key": "<key name>", "value": "<tag value>" } ]'


class Bucket(base.EnsurableAWSBaseActor):

    """Manage the state of a single S3 Bucket.

    The actor has the following functionality:

      * Ensure that an S3 bucket is present or absent.
      * Manage the bucket policy.
      * Manage the bucket Lifecycle configurations.
      * Enable or Suspend Bucket Versioning.
        Note: It is impossible to actually _disable_ bucket versioning -- once
        it is enabled, you can only suspend it, or re-enable it.

    **Note about Buckets with Files**

    Amazon requires that an S3 bucket be empty in order to delete it. Although
    we could recursively search for all files in the bucket and then delete
    them, this is a wildly dangerous thing to do inside the confines of this
    actor. Instead, we raise an exception and alert the you to the fact that
    they need to delete the files themselves.

    **Options**

    :name:
      The name of the bucket to operate on

    :state:
      (str) Present or Absent. Default: "present"

    :lifecycle:
      (:py:class:`LifecycleConfig`, None)

      A list of individual Lifecycle configurations. Each dictionary includes
      keys for:

      * `id`
      * `status`
      * `filter` (or `prefix`, which is deprecated)

      and at least one of:

      * `transitions` (or `transition`, which is deprecated)
      * `noncurrent_version_transitions` (or `noncurrent_version_transition`)
      * `expiration`
      * `noncurrent_version_expiration`
      * `abort_incomplete_multipart_upload`

      If an empty list is supplied, or the list in any way does not match what
      is currently configured in Amazon, the appropriate changes will be made.

    :logging:
      (:py:class:`LoggingConfig`, None)

      If a dictionary is supplied (`{'target': 'logging_bucket', 'prefix':
      '/mylogs'}`), then we will configure bucket logging to the supplied
      bucket and prefix. If `prefix` is missing then no prefix will be used.

      If `target` is supplied as an empty string (`''`), then we will disable
      logging on the bucket. If `None` is supplied, we will not manage logging
      either way.

    :tags:
      (:py:class:`TaggingConfig`, None)

      A list of dictionaries with a `key` and `value` key. Defaults to an empty
      list, which means that if you manually add tags, they will be removed.

    :policy:
      (str, None) A JSON file with the bucket policy. Passing in a blank string
      will cause any policy to be deleted. Passing in None (or not passing it
      in at all) will cause Kingpin to ignore the policy for the bucket
      entirely. Default: None

    :public_access_block_configuration:
      (:py:class:`PublicAccessBlockConfig`, None)

      If a dictionary is supplied, then it must conform to the
      :py:class:`PublicAccessBlockConfig` type and include all of the Public
      Access Block Configuration parameters.

      If an empty dictionary is supplied, then Kingpin will explicitly remove
      any Public Access Block Configurations from the bucket.

      Finally, if None is supplied, Kingpin will ignore the checks entirely on
      this portion of the bucket configuration.

      Default: None

    :region:
      AWS region (or zone) name, such as us-east-1 or us-west-2

    :versioning:
      (bool, None): Whether or not to enable Versioning on the bucket. If
      "None", then we don't manage versioning either way. Default: None

    **Examples**

    .. code-block:: json

       {
         "actor": "aws.s3.Bucket",
         "options": {
           "name": "kingpin-integration-testing",
           "region": "us-west-2",
           "policy": "./examples/aws.s3/amazon_put.json",
           "lifecycle": [
              {
                "id": "main",
                "status": "Enabled",
                "filter": {
                    "prefix": "/"
                },
                "expiration": 30,
              }
           ],
           "logging": {
             "target": "logs.myco.com",
             "prefix": "/kingpin-integratin-testing"
           },
           "tags": [
             {"key": "my_key", "value": "billing-grp-1"},
           ],
           "versioning": true,
         }
       }


    **Dry Mode**

    Finds the bucket if it exists (or tells you it would create it). Describes
    each potential change it would make to the bucket depending on the
    configuration of the live bucket, and the options that were passed into the
    actor.

    Will gracefully fail and alert you if there are files in the bucket and you
    are trying to delete it.
    """

    desc = "S3 Bucket {name}"

    all_options = {
        'name': (str, REQUIRED, 'Name of the S3 Bucket'),
        'state': (STATE, 'present',
                  'Desired state of the bucket: present/absent'),
        'lifecycle': (LifecycleConfig, None,
                      'List of Lifecycle configurations.'),
        'logging': (LoggingConfig, None,
                    'Logging configuration for the bucket'),
        'public_access_block_configuration': (
            PublicAccessBlockConfig, None,
            'Public Access Block Configuration'),
        'tags': (TaggingConfig, None,
                 'Array of dicts with the key/value tags'),
        'policy': ((str, None), None,
                   'Path to the JSON policy file to apply to the bucket.'),
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2'),
        'versioning': ((bool, None), None,
                       ('Desired state of versioning on the bucket: '
                        'true/false')),
    }

    unmanaged_options = ['name', 'region']

    def __init__(self, *args, **kwargs):
        super(Bucket, self).__init__(*args, **kwargs)

        # If the policy is None, or '', we simply set it to self.policy. If its
        # anything else, we parse it.
        self.policy = self.option('policy')
        if self.option('policy'):
            self.policy = self._parse_policy_json(self.option('policy'))

        # If the Lifecycle config is anything but None, we parse it and
        # pre-build all of our Lifecycle/Rule/Expiration/Transition objects.
        self.lifecycle = self.option('lifecycle')
        if self.lifecycle is not None:
            self.lifecycle = self._generate_lifecycle(self.option('lifecycle'))

        # If the PublicAccessBlockConfiguration is anything but None, we parse
        # it and pre-build the rules.
        self.access_block = self.option('public_access_block_configuration')
        if self.access_block is not None:
            self.access_block = self._snake_to_camel(self.access_block)

        # Start out assuming the bucket doesn't exist. The _precache() method
        # will populate this with True if the bucket does exist.
        self._bucket_exists = False

    def _snake_to_camel(self, data):
        """Converts a snake_case dict to CamelCase.

        To keep our LifecycleConfig schema in-line with the rest of Kingpin, we
        use snake_case for all key values. This method converts the snake_case
        into CamelCase for final uploading to Amazons API where CamelCase is
        required.
        """
        if isinstance(data, list):
            return [self._snake_to_camel(v) for v in data]
        elif isinstance(data, dict):
            return dict(
                (camelize(k), self._snake_to_camel(v)) for k, v
                in data.items())
        else:
            return data

    def _generate_lifecycle(self, config):
        """Generates a Lifecycle Configuration object.

        Takes the supplied configuration (a list of dicts) and turns them into
        proper Boto Lifecycle Rules, then returns a Lifecycle configuration
        object with these rules.

        args:
            config: A dict that matches the :py:class:`LifecycleConfig` schema.

        returns:
            :py:class:`boto.s3.lifecycle.Lifecycle`
            None: If the supplied configuration is empty
        """
        self.log.debug('Generating boto.s3.lifecycle.Lifecycle config..')

        # Generate a fresh Lifecycle configuration object
        rules = []
        for c in config:
            self.log.debug('Generating lifecycle rule from foo: %s' % c)

            # Convert the snake_case into CamelCase.
            c = self._snake_to_camel(c)

            # Fully capitalize the ID field
            c['ID'] = c.pop('Id')

            # If the Prefix was supplied in the old style, convert it into
            # the proper format for Amazon.
            if 'Prefix' in c:
                c['Filter'] = {'Prefix': c.pop('Prefix')}

            # If the Tranisition was supplied in the old style, convert it into
            # the proper format for Amazon.
            if 'Transition' in c:
                c['Transitions'] = [c.pop('Transition')]

            # If the NoncurrentVersionTransition was supplied in the old style,
            # convert it into the proper format for Amazon.
            if 'NoncurrentVersionTransition' in c:
                c['NoncurrentVersionTransitions'] = [
                    c.pop('NoncurrentVersionTransition')]

            # If the Expiration was supplied in the old style as a string/int,
            # convert it into the proper format for Amazon.
            if 'Expiration' in c and not isinstance(c['Expiration'], dict):
                c['Expiration'] = {'Days': int(c.pop('Expiration'))}

            # Finally add our rule to the lifecycle object
            rules.append(c)

        return rules

    @gen.coroutine
    def _precache(self):
        # Store a quick reference to whether or not the bucket exists or not.
        # This allows the rest of the getter-methods to know whether or not the
        # bucket exists and not make bogus API calls when the bucket doesn't
        # exist.
        buckets = yield self.api_call(self.s3_conn.list_buckets)
        matching = [
            b for b in buckets['Buckets'] if b['Name'] == self.option('name')]
        if len(matching) == 1:
            self._bucket_exists = True

    @gen.coroutine
    def _get_state(self):
        if not self._bucket_exists:
            raise gen.Return('absent')

        raise gen.Return('present')

    @gen.coroutine
    def _set_state(self):
        if self.option('state') == 'absent':
            yield self._verify_can_delete_bucket()
            yield self._delete_bucket()
        else:
            yield self._create_bucket()

    @gen.coroutine
    @dry('Would have created the bucket')
    def _create_bucket(self):
        """Creates an S3 bucket if its missing.

        returns:
            <A boto.s3.Bucket object>
        """
        params = {
            'Bucket': self.option('name')
        }

        if self.option('region') != 'us-east-1':
            params['CreateBucketConfiguration'] = {
                'LocationConstraint': self.option('region')
            }

        self.log.info('Creating bucket')
        yield self.api_call(self.s3_conn.create_bucket, **params)

    @gen.coroutine
    def _verify_can_delete_bucket(self):
        # Find out if there are any files in the bucket before we go to delete
        # it. We cannot delete a bucket with files in it -- nor do we want to.
        bucket = self.option('name')
        keys = yield self.api_call(self.s3_conn.list_objects, Bucket=bucket)

        if 'Contents' not in keys:
            raise gen.Return()

        if len(keys['Contents']) > 0:
            raise exceptions.RecoverableActorFailure(
                'Cannot delete bucket with keys: %s files found' % len(keys))

    @gen.coroutine
    @dry('Would have deleted bucket')
    def _delete_bucket(self):
        bucket = self.option('name')
        try:
            self.log.info('Deleting bucket %s' % bucket)
            yield self.api_call(self.s3_conn.delete_bucket, Bucket=bucket)
        except ClientError as e:
            raise exceptions.RecoverableActorFailure(
                'Cannot delete bucket: %s' % str(e))

    @gen.coroutine
    def _get_policy(self):
        if not self._bucket_exists:
            raise gen.Return(None)

        try:
            raw = yield self.api_call(
                self.s3_conn.get_bucket_policy,
                Bucket=self.option('name'))
            exist = json.loads(raw['Policy'])
        except ClientError as e:
            if 'NoSuchBucketPolicy' in str(e):
                raise gen.Return('')
            raise

        raise gen.Return(exist)

    @gen.coroutine
    def _compare_policy(self):
        new = self.policy
        if self.policy is None:
            self.log.debug('Not managing policy')
            raise gen.Return(True)

        exist = yield self._get_policy()

        # Now, diff our new policy from the existing policy. If there is no
        # difference, then we bail out of the method.
        diff = utils.diff_dicts(exist, new)
        if not diff:
            self.log.debug('Bucket policy matches')
            raise gen.Return(True)

        # Now, print out the diff..
        self.log.info('Bucket policy differs from Amazons:')
        for line in diff.split('\n'):
            self.log.info('Diff: %s' % line)

        raise gen.Return(False)

    @gen.coroutine
    def _set_policy(self):
        if self.policy == '':
            yield self._delete_policy()
        else:
            yield self._push_policy()

    @gen.coroutine
    @dry('Would have pushed bucket policy')
    def _push_policy(self):
        self.log.info('Pushing bucket policy %s' % self.option('policy'))
        self.log.debug('Policy doc: %s' % self.policy)

        try:
            yield self.api_call(
                self.s3_conn.put_bucket_policy,
                Bucket=self.option('name'),
                Policy=json.dumps(self.policy))
        except ClientError as e:
            if 'MalformedPolicy' in str(e):
                raise base.InvalidPolicy(str(e))

            raise exceptions.RecoverableActorFailure(
                'An unexpected error occurred: %s' % e)

    @gen.coroutine
    @dry('Would delete bucket policy')
    def _delete_policy(self):
        self.log.info('Deleting bucket policy')
        yield self.api_call(
            self.s3_conn.delete_bucket_policy,
            Bucket=self.option('name'))

    @gen.coroutine
    def _get_logging(self):
        if not self._bucket_exists:
            raise gen.Return(None)

        data = yield self.api_call(
            self.s3_conn.get_bucket_logging, Bucket=self.option('name'))

        if 'LoggingEnabled' not in data:
            self.log.debug('Logging is disabled')
            raise gen.Return({
                'target': '',
                'prefix': ''})

        self.log.debug('Logging is set to s3://%s/%s' %
                       (data['LoggingEnabled']['TargetBucket'],
                        data['LoggingEnabled']['TargetPrefix']))
        raise gen.Return({
            'target': data['LoggingEnabled']['TargetBucket'],
            'prefix': data['LoggingEnabled']['TargetPrefix']})

    @gen.coroutine
    def _set_logging(self):
        desired = self.option('logging')

        if desired is None:
            self.log.debug('Not managing logging')
            raise gen.Return()

        # If desired is False, check the state, potentially disable it, and
        # then bail out. Note, we check explicitly for 'target' to be set to
        # ''. Setting it to None, or setting the entire logging config to None
        # should not destroy any existing logging configs.
        if desired['target'] == '':
            yield self._disable_logging()
            raise gen.Return()

        # If desired has a logging or prefix config, check each one and
        # validate that they are correct.
        yield self._enable_logging(**desired)

    @gen.coroutine
    @dry('Bucket logging would have been disabled')
    def _disable_logging(self):
        self.log.info('Deleting Bucket logging configuration')
        yield self.api_call(
            self.s3_conn.put_bucket_logging,
            Bucket=self.option('name'),
            BucketLoggingStatus={})

    @gen.coroutine
    @dry('Bucket logging config would be updated to {target}/{prefix}')
    def _enable_logging(self, target, prefix):
        """Enables logging on a bucket.

        args:
            target: Target S3 bucket
            prefix: Target S3 bucket prefix
        """
        target_str = 's3://%s/%s' % (target, prefix.lstrip('/'))
        self.log.info('Updating Bucket logging config to %s' % target_str)

        try:
            yield self.api_call(
                self.s3_conn.put_bucket_logging,
                Bucket=self.option('name'),
                BucketLoggingStatus={
                    'LoggingEnabled': {
                        'TargetBucket': target,
                        'TargetPrefix': prefix,
                    }
                })
        except ClientError as e:
            raise InvalidBucketConfig(str(e))

    @gen.coroutine
    def _get_versioning(self):
        if not self._bucket_exists:
            raise gen.Return(None)

        existing = yield self.api_call(
            self.s3_conn.get_bucket_versioning,
            Bucket=self.option('name'))

        if ('Status' not in existing or
                existing['Status'] == 'Suspended'):
            self.log.debug('Versioning is disabled/suspended')
            raise gen.Return(False)

        self.log.debug('Versioning is enabled')
        raise gen.Return(True)

    @gen.coroutine
    def _set_versioning(self):
        if self.option('versioning') is None:
            self.log.debug('Not managing versioning')
            raise gen.Return()

        if self.option('versioning') is False:
            yield self._put_versioning('Suspended')
        else:
            yield self._put_versioning('Enabled')

    @gen.coroutine
    @dry('Bucket versioning would set to: {0}')
    def _put_versioning(self, state):
        self.log.info('Setting bucket object versioning to: %s' % state)
        yield self.api_call(
            self.s3_conn.put_bucket_versioning,
            Bucket=self.option('name'),
            VersioningConfiguration={'Status': state})

    @gen.coroutine
    def _get_lifecycle(self):
        if not self._bucket_exists:
            raise gen.Return(None)

        try:
            raw = yield self.api_call(
                self.s3_conn.get_bucket_lifecycle_configuration,
                Bucket=self.option('name'))
        except ClientError as e:
            if 'NoSuchLifecycleConfiguration' in str(e):
                raise gen.Return([])
            raise

        raise gen.Return(raw['Rules'])

    @gen.coroutine
    def _compare_lifecycle(self):
        existing = yield self._get_lifecycle()
        new = self.lifecycle

        if new is None:
            self.log.debug('Not managing lifecycle')
            raise gen.Return(True)

        # Now sort through the existing Lifecycle configuration and the one
        # that we've built locally. If there are any differences, we're going
        # to push an all new config.
        diff = utils.diff_dicts(
            json.loads(jsonpickle.encode(existing)),
            json.loads(jsonpickle.encode(new)))

        if not diff:
            raise gen.Return(True)

        self.log.info('Lifecycle configurations do not match. Updating.')
        for line in diff.split('\n'):
            self.log.info('Diff: %s' % line)
        raise gen.Return(False)

    @gen.coroutine
    def _set_lifecycle(self):
        if self.lifecycle == []:
            yield self._delete_lifecycle()
        else:
            yield self._push_lifecycle()

    @gen.coroutine
    @dry('Would have deleted the existing lifecycle configuration')
    def _delete_lifecycle(self):
        self.log.info('Deleting the existing lifecycle configuration.')
        yield self.api_call(
            self.s3_conn.delete_bucket_lifecycle,
            Bucket=self.option('name'))

    @gen.coroutine
    @dry('Would have pushed a new lifecycle configuration')
    def _push_lifecycle(self):
        self.log.debug('Lifecycle config: %s' %
                       jsonpickle.encode(self.lifecycle))

        self.log.info('Updating the Bucket Lifecycle config')
        try:
            yield self.api_call(
                self.s3_conn.put_bucket_lifecycle_configuration,
                Bucket=self.option('name'),
                LifecycleConfiguration={'Rules': self.lifecycle})
        except (ParamValidationError, ClientError) as e:
            raise InvalidBucketConfig('Invalid Lifecycle Configuration: %s'
                                      % str(e))

    @gen.coroutine
    def _get_public_access_block_configuration(self):
        if not self._bucket_exists:
            raise gen.Return(None)

        try:
            raw = yield self.api_call(
                self.s3_conn.get_public_access_block,
                Bucket=self.option('name'))
        except ClientError as e:
            if 'NoSuchPublicAccessBlockConfiguration' in str(e):
                raise gen.Return([])
            raise

        raise gen.Return(raw['PublicAccessBlockConfiguration'])

    @gen.coroutine
    def _set_public_access_block_configuration(self):
        if self.access_block == {}:
            yield self._delete_public_access_block_configuration()
        else:
            yield self._push_public_access_block_configuration()

        raise gen.Return()

    @gen.coroutine
    @dry('Would have deleted the existing public access block config')
    def _delete_public_access_block_configuration(self):
        self.log.info('Deleting the existing public access block config.')
        yield self.api_call(
            self.s3_conn.delete_public_access_block,
            Bucket=self.option('name'))

    @gen.coroutine
    @dry('Would have pushed a new public access block config')
    def _push_public_access_block_configuration(self):
        self.log.debug('Public Access Block Config: %s' %
                       jsonpickle.encode(self.access_block))

        self.log.info('Updating the Bucket Public Access Block Config')
        try:
            yield self.api_call(
                self.s3_conn.put_public_access_block,
                Bucket=self.option('name'),
                PublicAccessBlockConfiguration=self.access_block)
        except (ParamValidationError, ClientError) as e:
            raise InvalidBucketConfig(
                'Invalid Public Access Block Config: %s' % str(e))

    @gen.coroutine
    def _compare_public_access_block_configuration(self):
        existing = yield self._get_public_access_block_configuration()
        new = self.access_block

        if new is None:
            self.log.debug('Not managing public access block config')
            raise gen.Return(True)

        # Now sort through the existing Lifecycle configuration and the one
        # that we've built locally. If there are any differences, we're going
        # to push an all new config.
        diff = utils.diff_dicts(
            json.loads(jsonpickle.encode(existing)),
            json.loads(jsonpickle.encode(new)))

        if not diff:
            raise gen.Return(True)

        self.log.info(
            'Public Access Block Configurations do not match. Updating.')
        for line in diff.split('\n'):
            self.log.info('Diff: %s' % line)
        raise gen.Return(False)

    @gen.coroutine
    def _get_tags(self):
        if self.option('tags') is None:
            raise gen.Return(None)

        if not self._bucket_exists:
            raise gen.Return(None)

        try:
            raw = yield self.api_call(
                self.s3_conn.get_bucket_tagging,
                Bucket=self.option('name'))
        except ClientError as e:
            if 'NoSuchTagSet' in str(e):
                raise gen.Return([])
            raise

        # The keys in the sets returned always are capitalized (Key, Value) ...
        # but our schema uses lower case. Lowercase all of the keys before
        # returning them so that they are compared properly.
        tagset = []
        for tag in raw['TagSet']:
            tag = {k.lower(): v for k, v in list(tag.items())}
            tagset.append(tag)

        raise gen.Return(tagset)

    @gen.coroutine
    def _compare_tags(self):
        new = self.option('tags')
        if new is None:
            self.log.debug('Not managing Tags')
            raise gen.Return(True)

        exist = yield self._get_tags()

        diff = utils.diff_dicts(exist, new)
        if not diff:
            self.log.debug('Bucket tags match')
            raise gen.Return(True)

        self.log.info('Bucket tags differs from Amazons:')
        for line in diff.split('\n'):
            self.log.info('Diff: %s' % line)

        raise gen.Return(False)

    @gen.coroutine
    @dry('Would have pushed tags')
    def _set_tags(self):
        tags = self.option('tags')

        if tags is None:
            self.log.debug('Not managing tags')
            raise gen.Return(None)

        tagset = self._snake_to_camel(self.option('tags'))
        self.log.info('Updating the Bucket Tags')
        yield self.api_call(
            self.s3_conn.put_bucket_tagging,
            Bucket=self.option('name'),
            Tagging={'TagSet': tagset})



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rollbar`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The Rollbar Actor allows you to post Deploy messages to Rollbar when you
execute a code deployment.

**Required Environment Variables**

:ROLLBAR_TOKEN:
  Rollbar API Token
"""

import logging
import os
import urllib.request
import urllib.parse
import urllib.error

from tornado import gen
from tornado import httpclient

from kingpin import utils
from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


API_CONTENT_TYPE = 'application/json'
API_URL = 'https://api.rollbar.com/api/1'
API_DEPLOY_PATH = '%s/deploy/' % API_URL
API_PROJECT_PATH = '%s/project/' % API_URL

TOKEN = os.getenv('ROLLBAR_TOKEN', None)


class RollbarBase(base.HTTPBaseActor):

    """Simple Rollbar Base Abstract Actor"""

    def __init__(self, *args, **kwargs):
        """Check required environment variables."""
        super(RollbarBase, self).__init__(*args, **kwargs)

        if not TOKEN:
            raise exceptions.InvalidCredentials(
                'Missing the "ROLLBAR_TOKEN" environment variable.')

        self._token = TOKEN

    def _build_potential_args(self, potential_args):
        """Builds a full set of arguments to pass to Rollbar.

        Appends the authentication token and a few other bits to the
        arguments supplied.

        Args:
            potential_Args: A hash of potential arguments.

        Returns:
            A larger hash of arguments.
        """
        potential_args['access_token'] = self._token
        return potential_args

    @gen.coroutine
    @utils.retry(excs=(httpclient.HTTPError), retries=3)
    def _fetch_wrapper(self, *args, **kwargs):
        """Wrap the superclass _fetch method to catch known Rollbar errors.

        https://rollbar.com/docs/api_overview/
        """
        try:
            res = yield self._fetch(*args, **kwargs)
        except httpclient.HTTPError as e:
            # These are HTTPErrors that we know about, and can log specific
            # error messages for.

            if e.code in (401, 403):
                raise exceptions.InvalidCredentials(
                    'The "ROLLBAR_TOKEN" is invalid')
            elif e.code == 422:
                raise exceptions.RecoverableActorFailure(
                    'Unprocessable Entity - the request was parseable (i.e. '
                    'valid JSON), but some parameters were missing or '
                    'otherwise invalid.')
            elif e.code == 429:
                raise exceptions.RecoverableActorFailure(
                    'Too Many Requests - If rate limiting is enabled for '
                    'your access token, this return code signifies that the '
                    'rate limit has been reached and the item was not '
                    'processed.')
            else:
                # We ran into a problem we can't handle. Also, keep in mind
                # that @utils.retry() was used, so this error happened several
                # times before getting here. Raise it.
                raise exceptions.RecoverableActorFailure(
                    'Unexpected error from Rollbar API: %s' % e)

        raise gen.Return(res)

    @gen.coroutine
    def _project(self):
        """Get a project description back from Rollbar.

        This method is used as a simple test that the API keys work. It access
        the list of projects from Rollbar and raises the appropriate exceptions
        if it cannot.

        https://rollbar.com/docs/api/projects/#list-your-projects

        Raises:
            gen.Return(<Dictionary of the response from Rollbar>)
        """

        args = self._build_potential_args({})
        url = self._generate_escaped_url(API_PROJECT_PATH, args)
        res = yield self._fetch_wrapper(url)
        raise gen.Return(res)


class Deploy(RollbarBase):

    """Posts a Deploy message to Rollbar.

    https://rollbar.com/docs/deploys_other/

    **API Token**

    You must use an API token created in your *Project Access Tokens* account
    settings section. This token should have *post_server_item* permissions for
    the actual deploy, and *read* permissions for the Dry run.

    **Options**

    :environment:
      The environment to deploy to

    :revision:
      The deployment revision

    :local_username:
      The user who initiated the deploy

    :rollbar_username:
      *(Optional)* The Rollbar Username to assign the deploy to

    :comment:
      *(Optional)* Comment describing the deploy

    **Examples**

    .. code-block:: json

       { "actor": "rollbar.Deploy",
         "desc": "update rollbar deploy",
         "options": {
           "environment": "Prod",
           "revision": "%DEPLOY%",
           "local_username": "Kingpin",
           "rollbar_username": "Kingpin",
           "comment": "some comment %DEPLOY%"
         }
       }

    **Dry Mode**

    Accesses the Rollbar API and validates that the token can access your
    project.
    """
    all_options = {
        'environment': (str, REQUIRED, 'Name of the environment to deploy'),
        'revision': (str, REQUIRED, 'Revision number/sha being deployed'),
        'local_username': (str, 'Kingpin', 'User who deployed'),
        'rollbar_username': (str, '', 'Rollbar username'),
        'comment': (str, '', 'Deploy comment')
    }

    desc = "Sending Deploy {environment}/{revision}"

    @gen.coroutine
    def _deploy(self):
        """Posts a Deploy to rollbar.

        https://rollbar.com/docs/deploys_other/

        Raises:
            gen.Return(<Dictionary of the response from Rollbar>)
        """

        rollbar_username = self.option('rollbar_username')
        if rollbar_username == '':
            rollbar_username = None

        args = self._build_potential_args({
            'environment': self.option('environment'),
            'revision': self.option('revision'),
            'local_username': self.option('local_username'),
            'rollbar_username': rollbar_username,
            'comment': self.option('comment')
        })

        escaped_post = urllib.parse.urlencode(args)
        res = yield self._fetch_wrapper(API_DEPLOY_PATH, post=escaped_post)
        raise gen.Return(res)

    @gen.coroutine
    def _execute(self):
        """Executes an actor and yields the results when its finished.

        raises: gen.Return()
        """
        rollbar_string = (
            'Rollbar Deploy Notification %s/%s' %
            (self.option('environment'), self.option('revision')))

        if self._dry:
            self.log.info('Would have sent %s, but instead just validating '
                          'API key.' % rollbar_string)
            yield self._project()
            raise gen.Return()

        self.log.info('Sending %s' % rollbar_string)
        yield self._deploy()
        raise gen.Return()



# Copyright 2014 Nextdoor.com, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import shutil

from distutils.command.clean import clean
from distutils.command.sdist import sdist
from setuptools import Command
from setuptools import setup
from setuptools import find_packages

from kingpin.version import __version__

PACKAGE = 'kingpin'
DIR = os.path.dirname(os.path.realpath(__file__))


def maybe_rm(path):
    """Simple method for removing a file/dir if it exists"""
    if os.path.exists(path):
        try:
            shutil.rmtree(path)
        except:
            os.remove(path)


def read_requirements(requirements):
    """Parse requirements from requirements.txt."""
    reqs_path = requirements
    with open(reqs_path, 'r') as f:
        requirements = [line.rstrip() for line in f]
    return requirements


class Pep8Command(Command):
    description = 'Pep8 Lint Checks'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # Don't import the pep8 module until now because setup.py needs to be
        # able to install Pep8 if its missing.
        import pep8
        pep8style = pep8.StyleGuide(parse_argv=True, config_file='pep8.cfg')
        report = pep8style.check_files([PACKAGE])
        if report.total_errors:
            sys.exit('ERROR: Pep8 failed with exit %d errors' %
                     report.total_errors)


class PyflakesCommand(Command):
    description = 'Pyflakes Checks'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # Don't import the pyflakes code until now because setup.py needs to be
        # able to install Pyflakes if its missing. This localizes the import to
        # only after the setuptools code has run and verified everything is
        # installed.
        from pyflakes import api
        from pyflakes import reporter

        # Run the Pyflakes check against our package and check its output
        val = api.checkRecursive([PACKAGE], reporter._makeDefaultReporter())
        if val > 0:
            sys.exit('ERROR: Pyflakes failed with exit code %d' % val)


class UnitTestCommand(Command):
    description = 'Run unit tests'
    user_options = []
    args = [PACKAGE,
            '--cover-erase',
            '--cover-min-percentage=100',
            '--with-coverage',
            '--cover-package=%s' % PACKAGE,
            '-v']

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # (imported here so that the setup.py can install the rest of the test
        # requirements, including nose...)
        import nose
        maybe_rm('.coverage')
        val = nose.run(argv=self.args)

        if not val:
            sys.exit('ERROR:Tests failed')


class IntegrationTestCommand(UnitTestCommand):
    description = 'Run full integration tests and unit tests'
    tests = os.environ.get('INTEGRATION_TESTS', 'integration')

    args = [PACKAGE,
            '--with-coverage',
            '--cover-package=%s' % PACKAGE,
            '-v',
            '--include=integration',
            ]

    for test in tests.split(','):
        args.append('--attr=%s' % test)


class CleanHook(clean):

    def run(self):
        clean.run(self)

        maybe_rm('%s.egg-info' % PACKAGE)
        maybe_rm('dist')
        maybe_rm('.coverage')
        maybe_rm('version.rst')
        maybe_rm('MANIFEST')


class SourceDistHook(sdist):

    def run(self):
        with open('version.rst', 'w') as f:
            f.write(':Version: %s\n' % __version__)
        shutil.copy('%s/README.rst' % DIR, 'README')
        sdist.run(self)
        os.unlink('MANIFEST')
        os.unlink('README')
        os.unlink('version.rst')


setup(
    name=PACKAGE,
    version=__version__,
    description='Deployment Automation Engine',
    long_description=open('%s/README.rst' % DIR).read(),
    author='Nextdoor Engineering',
    author_email='eng@nextdoor.com',
    url='https://github.com/Nextdoor/kingpin',
    download_url='http://pypi.python.org/pypi/%s#downloads' % PACKAGE,
    license='Apache License, Version 2.0',
    keywords='apache',
    packages=find_packages(),
    test_suite='nose.collector',
    tests_require=read_requirements('%s/requirements.test.txt' % DIR),
    python_requires='~=3.7',
    setup_requires=[],
    install_requires=read_requirements('%s/requirements.txt' % DIR),
    entry_points={
        'console_scripts': [
            'kingpin = kingpin.bin.deploy:begin'
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Software Development',
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Operating System :: POSIX',
        'Natural Language :: English',
    ],
    cmdclass={
        'sdist': SourceDistHook,
        'clean': CleanHook,
        'pep8': Pep8Command,
        'pyflakes': PyflakesCommand,
        'integration': IntegrationTestCommand,
        'test': UnitTestCommand,
    },
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.utils`
^^^^^^^^^^^^^^^^^^^^^^^^^^^

Misc methods for dealing with Actors.
"""

import logging
import time

from tornado import gen

from kingpin import utils
from kingpin.actors import exceptions

log = logging.getLogger(__name__)


__author__ = 'Matt Wise <matt@nextdoor.com>'


def dry(dry_message):
    """Coroutine-compatible decorator to dry-run a method.

    Note: this must act on a :py:mod:`~kingpin.actors.base.BaseActor` object.

    Example usage as decorator:

        >>> @gen.coroutine
        ... @dry('Would have done that {thing}')
        ... def do_thing(self, thing):
        ...     yield api.do_thing(thing)
        ...
        >>> yield do_thing(thing="yeah man, that thing")

    Args:
        dry_message: The message to print out instead of doing the actual
        function call. This string is passed through format(kwargs), so any
        variables you'd like can be substituted as long as they're passed to
        the method being wrapped.
    """
    # TODO: Bring these back when we have log.trace
    # log.debug('Creating _skip_on_dry decorator with "%s"' % dry_message)

    def _skip_on_dry(f):
        # TODO: Bring these back when we have log.trace
        # log.debug('Decorating function "%s" with _skip_on_dry' % f)

        def wrapper(self, *args, **kwargs):
            # _Always_ compile the message we'd use in the event of a Dry run.
            # This ensures that our test cases catch any time invalid **kwargs
            # are passed in.
            msg = dry_message.format(*args, **kwargs)

            if self._dry:
                self.log.warning(msg)
                raise gen.Return()
            ret = yield gen.coroutine(f)(self, *args, **kwargs)
            raise gen.Return(ret)

        return wrapper
    return _skip_on_dry


def timer(f):
    """Coroutine-compatible function timer.

    Records statistics about how long a given function took, and logs them
    out in debug statements. Used primarily for tracking Actor execute()
    methods, but can be used elsewhere as well.

    Note: this must act on a :py:mod:`~kingpin.actors.base.BaseActor` object.

    Example usage:
        >>> @gen.coroutine
        ... @timer()
        ... def execute(self):
        ...     raise gen.Return()
    """

    def _wrap_in_timer(self, *args, **kwargs):
        # Log the start time
        start_time = time.time()

        # Begin the execution
        ret = yield gen.coroutine(f)(self, *args, **kwargs)

        # Log the finished execution time
        exec_time = "%.2f" % (time.time() - start_time)
        self.log.debug('%s.%s() execution time: %ss' %
                       (self._type, f.__name__, exec_time))

        raise gen.Return(ret)
    return _wrap_in_timer


def get_actor(config, dry):
    """Returns an initialized Actor object.

    Args:
        config: A dictionary of configuration data that conforms to our v1
                schema in kingpin.schema. Looks like this:

                {
                 'desc': <string description of actor>,
                 'actor': <string name of actor>
                 'options': <dict of options to pass to actor>
                 'warn_on_failure': <bool>
                 'condition': <string or bool>
                 }

        dry: Boolean whether or not in Dry mode
        warn_on_failure: Boolean

    Returns:
        <actor object>
    """
    # Copy the supplied dict before we modify it below
    config = dict(config)

    # Get the name of the actor, and pull it out of the config because its
    # not a valid kwarg for an Actor object.
    actor_string = config.pop('actor')

    # Create a copy of the config dict, but strip out the tokens. They likely
    # contain credentials! This is used purely for this debug message below.
    #
    # Known actors that do this are misc.Macro, group.Sync, group.Async
    clean_config = config.copy()
    clean_config['init_tokens'] = '<hidden>'

    log.debug('Building Actor "%s" with args: %s' %
              (actor_string, clean_config))
    ActorClass = get_actor_class(actor_string)
    return ActorClass(dry=dry, **config)


def get_actor_class(actor):
    """Returns a Class Reference to an Actor by string name.

    Args:
        actor: String name of the actor to find.

    Returns:
        <Class Ref to Actor>
    """
    expected_exceptions = (AttributeError, ImportError, TypeError)

    # Try to load our local actors up first. Assume that the
    # 'kingpin.actors.' prefix was not included in the name.
    for prefix in ['kingpin.actors.', '', 'actors.']:
        full_actor = prefix + actor
        try:
            return utils.str_to_class(full_actor)
        except expected_exceptions as e:
            log.debug('Tried importing "%s" but failed: %s' % (full_actor, e))

    msg = 'Unable to import "%s" as a valid Actor.' % actor
    raise exceptions.InvalidActor(msg)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.mci`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. _MultiCloudImages_:
    http://reference.rightscale.com/api1.5/resources/ResourceMultiCloudImages.html
    http://reference.rightscale.com/api1.5/resources/ResourceMultiCloudImageSettings.html
"""

import logging
import mock

from tornado import gen

from kingpin.actors import exceptions
from kingpin.actors.utils import dry
from kingpin.actors.rightscale import base
from kingpin.constants import SchemaCompareBase
from kingpin.constants import REQUIRED
from kingpin.constants import STATE

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


class MultiCloudImageSettings(SchemaCompareBase):

    """Provides JSON-Schema based validation of the supplied MultiCloudImage.

    Each image is a dictionary that must have a cloud, image, instance_type and
    optionally some user_data.

    .. code-block:: json

        [
          {
              "cloud": "EC2 us-west-2",
              "image": "ami-e29774d1",
              "instance_type": "m1.small",
              "user_data": "cd /bin/bash"
          },
          {
              "cloud": "EC2 us-west-1",
              "image": "ami-b58142f1",
              "instance_type": "m1.small",
              "user_data": "cd /bin/bash"
          }
        ]
    }
    """

    SCHEMA = {
        'type': ['array', 'null'],
        'uniqueItems': True,
        'items': {
            'type': 'object',
            'required': ['cloud', 'image', 'instance_type'],
            'additionalProperties': False,
            'properties': {
                'cloud': {
                    'type': 'string',
                    'minLength': 1,
                    'maxLength': 255,
                },
                'image': {
                    'type': 'string',
                },
                'instance_type': {
                    'type': 'string',
                },
                'user_data': {
                },
            }
        }
    }


class MCIBaseActor(base.RightScaleBaseActor):

    """Abstract MCI Actor that provides some utility methods."""

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""

        super(MCIBaseActor, self).__init__(*args, **kwargs)

        # Self.changed will be set to True if any _change_ occurs to an MCI
        # while its being managed. This is later leveraged by the 'commit'
        # option.
        self.changed = False

    @gen.coroutine
    def _get_mci(self, name):
        """Searches RightScale for an MCI by name.

        Wrapper around our find_by_name_and_keys() mechanism so that we return
        either the proper MCI, or None if one isn't found.

        args:
            name: MCI name to search for
        """
        self.log.debug('Searching for MCI "%s"' % name)
        mci = yield self._client.find_by_name_and_keys(
            collection=self._client._client.multi_cloud_images,
            name=name,
            revision=0)

        # Default searches return us an empty list if there are no matching
        # resources, or return us the exact resource we're looking for. Thus,
        # if the return value is not a list, then we know we got the MCI back.
        # got a list bac
        if not isinstance(mci, list):
            raise gen.Return(mci)

        # If we got a list back, return None because that means that the MCI
        # doesn't already exist.
        if isinstance(mci, list) and len(mci) == 0:
            raise gen.Return(None)

        # On anything else, raise an exception. Something really strange
        # happened.
        raise exceptions.RecoverableActorFailure(
            'Found too many matching MCI images with the same name, '
            'this shouldn\'t be possible ... so something bad has happened.')

    @gen.coroutine
    def _create_mci(self, name, params):
        """Creates a RightScale MCI if it doesn't exist.

        See http://reference.rightscale.com/api1.5/
            resources/ResourceMultiCloudImages.html#create

        Note, returns a mocked out MCI object if we're in a DRY run. This
        allows the rest of the actors to pretend like an MCI was created and
        continue with their checks.

        args:
            name: The name of the MCI we're creating
            params: RightScale-compatible list of tuples with the required
                    parameters defined in the URL above.

        returns:
            <rightscale.multi_cloud_image> object
        """
        mci = yield self._get_mci(name)
        if mci:
            raise gen.Return(mci)

        # If we're in a dry run, we return back a mocked out MCI image object
        # because its passed around a bunch for the other methods.
        if self._dry:
            self.log.warning('Would have created MCI: %s' % name)

            mci = mock.MagicMock(name=name)
            mci.href = None
            mci.soul = {
                'name': '<mocked MCI %s>' % name,
                'description': None
            }
            raise gen.Return(mci)

        self.log.info('Creating MCI %s' % self.option('name'))
        mci = yield self._client.create_resource(
            self._client._client.multi_cloud_images, params)
        self.changed = True
        raise gen.Return(mci)

    @gen.coroutine
    @dry('Would have deleted MCI: {name}')
    def _delete_mci(self, name):
        """Deletes a RightScale MCI if it exists.

        See http://reference.rightscale.com/api1.5/
            resources/ResourceMultiCloudImages.html#destroy

        args:
            name: The name of the MCI we're creating
        """
        mci = yield self._get_mci(name)
        if not mci:
            raise gen.Return()

        self.log.info('Deleting MCI %s' % name)
        yield self._client.destroy_resource(mci)
        self.changed = True

    @gen.coroutine
    def _get_mci_setting_def(self, settings):
        """Returns a fully populated set of Multi Cloud Image Settings.

        See http://reference.rightscale.com/api1.5/
        resources/ResourceMultiCloudImageSettings.html for details.

        This method takes in a dictionary with as set of parameters (cloud,
        image, instance_type, user_data) and returns a fully ready-to-use
        set of RightScale-formatted parameters to update that image. The method
        handles discovering the RightScale HREFs for the cloud, image and
        instance_type options.

        Args:
            settings: A dictionary with the keys: cloud, image,
            instance_type, user_data

        Returns:
            A RightScale-formatted array of tuples.
        """

        # Get our cloud object first -- its required so that we can search for
        # the image/ramdisk/etc hrefs.
        cloud = yield self._client.find_by_name_and_keys(
            collection=self._client._client.clouds,
            name=settings['cloud'])
        if not cloud:
            raise exceptions.InvalidOptions(
                'Invalid Cloud name supplied: %s' % settings['cloud'])

        # Find our image by searching for the resource_uid that matches.
        image = yield self._client.find_by_name_and_keys(
            collection=cloud.images,
            resource_uid=settings['image'])
        if not image:
            raise exceptions.InvalidOptions(
                'Invalid cloud image name supplied: %s' % settings['image'])

        # Find our instance type now too
        instance = yield self._client.find_by_name_and_keys(
            collection=cloud.instance_types,
            name=settings['instance_type'])
        if not instance:
            raise exceptions.InvalidOptions(
                'Invalid cloud instance_type supplied: %s' %
                settings['instance_type'])

        # Generate our mci parameters, and each of the image settings
        # parameters. This validates that our inputs are all correct one last
        # time.
        params = {'cloud_href': cloud.href,
                  'image_href': image.href,
                  'instance_type_href': instance.href,
                  }
        if settings.get('user_data'):
            params['user_data'] = settings['user_data']

        definition = self._generate_rightscale_params(
            prefix='multi_cloud_image_setting',
            params=params)

        self.log.debug('Prepared MCI Image Definition: %s' % definition)
        raise gen.Return(definition)

    @gen.coroutine
    @dry('Would have added {cloud} to {mci.soul[name]}')
    def _create_mci_setting(self, cloud, mci, params):
        self.log.info('Adding Cloud %s to MCI %s' % (cloud, mci.soul['name']))
        yield self._client.create_resource(mci.settings, params)
        self.changed = True

    @gen.coroutine
    @dry('Would have updated the {mci_setting.links[cloud]} image settings')
    def _update_mci_setting(self, mci_setting, params):
        self.log.info('Updating Cloud %s settings' %
                      mci_setting.links['cloud'])
        yield self._client.update(mci_setting, params)
        self.changed = True

    @gen.coroutine
    @dry('Would have deleted the {mci_setting.links[cloud]} image settings')
    def _delete_mci_setting(self, mci_setting):
        self.log.info('Deleting Cloud %s settings' %
                      mci_setting.links['cloud'])
        yield self._client.destroy_resource(mci_setting)
        self.changed = True

    @gen.coroutine
    @dry('Would have updated the MCI description to: {description}')
    def _update_description(self, mci, description, params):
        self.log.info('Updating MCI description: %s' % description)
        mci = yield self._client.update(mci, params)
        self.changed = True
        raise gen.Return(mci)

    def _diff_setting(self, mci_setting, new_params):
        """Compares MCI Cloud Image settings to the newly requested params.

        Takes in a rightscale.multi_cloud_image.settings object and converts it
        into a list of RightScale-compatible tuples. Then compares these tuples
        to the tuples we want the cloud image setting to have. If they match,
        returns True, otherwise False.

        Note, strips out the user_data that the user supplied from the
        new_params because the user_data cannot be pulled down from the API at
        this time.

        args:
            mci_setting: A rightscale.multi_cloud_image.settings object
            new_params: A list of RightScale API-compatble tuples that we are
            comparing against.

        returns:
            boolean
        """
        links = mci_setting.links
        existing_params = [
            ('multi_cloud_image_setting[image_href]', links['image']),
            # See note in _diff_setting() about why this is commented out
            # ('multi_cloud_image_setting[user_data]', data[...]),
            ('multi_cloud_image_setting[instance_type_href]',
             links['instance_type']),
            ('multi_cloud_image_setting[cloud_href]', links['cloud'])]

        # NOTE: the user_data cannot be diffed because RightScale does not
        # provide a way to get this back via their API. This sucks, but for the
        # purposes of this diff_setting() method, we explicitly strip out the
        # user_data and only diff the other settings.
        new_params = [
            (key, value) for key, value in new_params if key !=
            'multi_cloud_image_setting[user_data]']

        return set(existing_params) != set(new_params)


class MCI(MCIBaseActor):

    """Manages the state of an RightScale MultiCloudImage.

    This actor is able to create, destroy, modify and commit MultiCloudImage
    revisions.

    Options match the documentation in RightScale:
    http://reference.rightscale.com/api1.5/resources/ResourceMultiCloudImages.html

    **Committing**

    If you wish to commit the MCI, set the `commit` option to the commit
    message you wish to use. If any change is made to the resource, Kingpin
    will commit a new revision with this message.

    **Options**

    :name:
      (str) The name of the MCI to be updated.

    :state:
      (str) Present or Absent. Default: "present"

    :commit:
      (str, None) If present, upon making changes to the HEAD version of the
      MCI, we will commit a new revision to RightScale. The provided string
      will be used as the commit message. Default: None

    :tags:
      (list, str) A list of tags to add to the MultiCloud image.
      Default: None

    :description:
      (str) The description of the MCI image itself.
      Default: ""

    :images:
       A list of dicts that each describe a single cloud and the image in that
       cloud to launch. See below for details.

    **Image Definitions**

    Each cloud image definition is a dictionary that takes a few keys.

    :cloud:
      (str) The name of the cloud as found in RightScale. We use the cloud
      'Name' which can be found in your `Settings -> Account Settings -> Clouds
      -> insert_cloud_here` page. For example `AWS us-west-2`.

    :image:
      (str) The cloud-specific Image UID. For example `ami-a1234abc`.

    :instance_type:
      (str) The default instance type to launch when this AMI is launched. For
      example, `m1.small`.
      (*optional*)

    :user_data:
      (str) The custom user data to pass to the instance on-bootup.
      (*optional*)

    **Examples**

    .. code-block:: json

        { "actor": "rightscale.MCI",
          "options": {
              "name": "Ubuntu i386 14.04",
              "description": "this is our test mci",
              "images": [
                {
                    "cloud": "EC2 us-west-2",
                    "image": "ami-e29774d1",
                    "instance_type": "m1.small",
                    "user_data": "cd /bin/bash"
                },
                {
                    "cloud": "EC2 us-west-1",
                    "image": "ami-b58142f1",
                    "instance_type": "m1.small",
                    "user_data": "cd /bin/bash"
                }
              ]
          }
        }

    """

    all_options = {
        'name': (str, REQUIRED, 'The name of the MCI to be updated'),
        'state': (STATE, 'present',
                  'Desired state of the image: present/absent'),
        'commit': (str, None, 'Commit a new MCI revision if changes made'),
        'description': (str, None, 'The description of the MCI to be updated'),
        'tags': ((list, str), None, 'List of tags to apply to the MCI'),
        'images': (
            MultiCloudImageSettings, None,
            'A list of objects that describe our per cloud image settings'),
    }

    desc = 'RightScale MCI {name}'

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""

        super(MCI, self).__init__(*args, **kwargs)
        self.changed = False
        self._mci_params = self._generate_rightscale_params(
            prefix='multi_cloud_image',
            params={
                'description': self.option('description'),
                'name': self.option('name')
            })

    @gen.coroutine
    def _ensure_mci(self):
        state = self.option('state')
        name = self.option('name')
        self.log.info('Ensuring that MCI %s is %s' % (name, state))
        mci = yield self._get_mci(name)

        if state == 'absent' and mci is None:
            self.log.debug('MCI does not exist')
        elif state == 'absent' and mci:
            yield self._delete_mci(name=name)
            mci = None
        elif state == 'present' and mci is None:
            mci = yield self._create_mci(name=name,
                                         params=self._mci_params)
        elif state == 'present' and mci:
            self.log.debug('MCI exists')

        raise gen.Return(mci)

    @gen.coroutine
    def _ensure_description(self, mci):
        existing = mci.soul['description']
        new = self.option('description')

        if existing == new:
            self.log.debug('Descriptions match')
            raise gen.Return()

        yield self._update_description(
            mci, description=new, params=self._mci_params)

    @gen.coroutine
    @dry('Would have force-set the user_data')
    def _force_mci_setting_user_data(self, mci_setting, params):
        """Temporary method to forcefully set the user_data on an MCI.

        RightScale does not currently allow us to check what the existing
        user_data is for a given MCI Setting. For that reason, we have to
        forcefully reset this data every time the code runs to ensure that its
        as expected.

        We treat this as a separate method call because we hope to remove this
        in the future, and keeping it separate makes it eaiser to remove down
        the road. Additionally, this method does not trigger the self.changed
        boolean, so it does not cause a new committed revision of the MCI to be
        created on every run.

        Again, we hope to destroy this method soon!

        args:
            mci_setting: rightscale.multi_cloud_images.setting object
            params: The parameters to force-set for the MCI setting
        """
        self.log.info('Force-setting the user_data')
        try:
            yield self._client.update(mci_setting, params)
        except StopIteration:
            return

    @gen.coroutine
    def _ensure_settings(self, mci):
        # Get the existing MCI cloud settings
        existing = yield self._client.show(mci.settings)

        # Go off and generate what the new settings should look like -- this is
        # an async IO operation because we let the users give us sane
        # human-readable resource names, and then we go and discover their raw
        # HREFs in the RightScale API.
        tasks = []
        for image in self.option('images'):
            tasks.append(self._get_mci_setting_def(image))
        new = yield tasks

        # First, lets create-or-update anything thats not in the
        # existing list of configured settings.
        tasks = []
        for new_setting in new:
            # Dive into the list of tuples for this cloud image setting and
            # find its HREF.
            new_cloud_href = [
                s[1] for s in new_setting if 'cloud_href' in s[0]][0]

            # Dig through the existing settings list and look for settings
            # objects that match the cloud name of the new_setting dict. If
            # none are received, this is an empty list.. otherwise, it should
            # only have one item in it.
            existing_setting = [
                s for s in existing if s.links['cloud'] == new_cloud_href]

            # If the configured image doesn't exist in our existing list of
            # cloud images, then lets just create it.
            if len(existing_setting) == 0:
                tasks.append(self._create_mci_setting(
                    cloud=new_cloud_href,
                    mci=mci,
                    params=new_setting))
                continue

            # Now, if the cloud IS defined already in the MCi, then we have to
            # see if they match .. or if they don't, we'll have to update them.
            #
            # Note, it should be impossible according to RightScale for this to
            # be > 1 ... you cannot have multiple cloud images for the same
            # cloud href, so we only need to think about the scenario where
            # this is == 1.
            if len(existing_setting) == 1:
                if self._diff_setting(existing_setting[0], new_setting):
                    tasks.append(self._update_mci_setting(
                        mci_setting=existing_setting[0],
                        params=new_setting))

                # Temporary -- when rightscale lets us check the user_data
                # value via the api, we won't have to do this anymore.
                tasks.append(self._force_mci_setting_user_data(
                    mci_setting=existing_setting[0],
                    params=new_setting))

        # Now that we've added or updated the cloud images we _want_, lets
        # purge any that are no longer listed.
        for existing_setting in existing:
            existing_cloud_href = existing_setting.links['cloud']
            new_cloud_hrefs = [
                dict(s)['multi_cloud_image_setting[cloud_href]'] for s in new]

            if existing_cloud_href not in new_cloud_hrefs:
                tasks.append(self._delete_mci_setting(
                    mci_setting=existing_setting))

        yield tasks

    @gen.coroutine
    @dry('Would have committed HEAD to a revision')
    def _commit(self, mci, message):
        self.log.info('Committing a new revision')
        ret = yield self._client.commit_resource(
            res=mci,
            res_type=self._client._client.multi_cloud_images,
            message=message)
        self.log.info('Committed revision %s' % ret.soul['revision'])

    @gen.coroutine
    def _execute(self):

        mci = yield self._ensure_mci()

        # If we're deleting the MCI, then there is no need to continue after
        # we've done that.
        if self.option('state') == 'absent':
            raise gen.Return()

        # Ensure that the description is up to date
        yield self._ensure_description(mci)

        # If tags were supplied, then manage them.
        if self.option('tags'):
            yield self._ensure_tags(mci, self.option('tags'))

        # Ensure that all of the configured images themselves match
        yield self._ensure_settings(mci)

        # Finally, if we're committing and a change was made, commit!
        if self.changed and self.option('commit'):
            yield self._commit(mci, self.option('commit'))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2013 Nextdoor.com, Inc.

"""
:mod:`kingpin.utils`
^^^^^^^^^^^^^^^^^^^^

Common package for utility functions.
"""

from logging import handlers
import difflib
import datetime
import demjson
import functools
import importlib
import logging
import os
import pprint
import re
import sys
import yaml
import io
from io import IOBase

from tornado import gen
from tornado import ioloop
import http.client
import rainbow_logging_handler

from kingpin import exceptions


__author__ = 'Matt Wise (matt@nextdoor.com)'

log = logging.getLogger(__name__)

# Constants for some of the utilities below
STATIC_PATH_NAME = 'static'

# Disable the global threadpool defined here to try to narrow down the random
# unit test failures regarding the IOError. Instead, instantiating a new
# threadpool object for every thread using the 'with' context below.
#
# # Allow up to 10 threads to be executed at once. This is arbitrary, but we
# # want to prvent the app from going thread-crazy.
# THREADPOOL_SIZE = 10
# THREADPOOL = futures.ThreadPoolExecutor(THREADPOOL_SIZE)


def str_to_class(string):
    """Method that converts a string name into a usable Class name

    This is used to take the 'actor' value from the JSON object and convert it
    into a valid object reference.

    Args:
        cls: String name of the wanted class and package.
             eg: kingpin.actors.foo.bar
             eg: misc.Sleep
             eg: actors.misc.Sleep
             eg: my.private.Actor

    Returns:
        A reference to the actual Class to be instantiated
    """
    # Split the string up. The last element is the Class, the rest is
    # the package name.
    string_elements = string.split('.')
    class_name = string_elements.pop()
    module_name = '.'.join(string_elements)

    m = importlib.import_module(module_name)
    return getattr(m, class_name)


def setup_root_logger(level='warn', syslog=None, color=False):
    """Configures the root logger.

    Args:
        level: Logging level string ('warn' is default)
        syslog: String representing syslog facility to output to.  If empty,
        logs are written to console.
        color: Colorize the log output

    Returns:
        A root Logger object
    """

    # Get the logging level string -> object
    level = 'logging.%s' % level.upper()
    level_obj = str_to_class(level)

    # Get our logger
    logger = logging.getLogger()

    # Set our default logging level
    logger.setLevel(level_obj)

    # Set the default logging handler to stream to console..
    if color:
        # Patch the handler's 'is_tty()' method to return True. If the user
        # asked for color, we give them color. The is_tty() method calls the
        # sys.stdout.isatty() method and then refuses to give color output on
        # platforms like Jenkins, where this code is likely to be run.
        rainbow_logging_handler.RainbowLoggingHandler.is_tty = True

        handler = rainbow_logging_handler.RainbowLoggingHandler(
            sys.stdout,

            # Disable colorization of the 'info' log statements. If the code is
            # run in an environment like Jenkins, the background is white, and
            # we don't want to force these log lines to be white as well.
            color_message_info=(None, None, False)
        )
    else:
        handler = logging.StreamHandler()

    # Get our PID .. used below in the log line format.
    details = ''
    if level_obj <= 10:
        details = str(os.getpid()) + ' [%(name)-40s] [%(funcName)-20s]'

    # If syslog enabled, then override the logging handler to go to syslog.
    asctime = '%(asctime)-10s '
    if syslog is not None:
        asctime = ''
        handler = handlers.SysLogHandler(address=('127.0.0.1', 514),
                                         facility=syslog)

    fmt = asctime + '%(levelname)-8s ' + details + ' %(message)s'
    formatter = logging.Formatter(fmt)

    # Append the formatter to the handler, then set the handler as our default
    # handler for the root logger.
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger


def super_httplib_debug_logging():
    """Enables DEBUG logging deep in HTTPLIB.

    HTTPLib by default doens't log out things like the raw HTTP headers,
    cookies, response body, etc -- even when your main logger is in DEBUG mode.
    This is because its a security risk, as well as just highly verbose.

    For the purposes of debugging though, this can be useful. This method
    enables deep debug logging of the HTTPLib web requests. This is highly
    insecure, but very useful when troubleshooting failures with remote API
    endpoints.

    Returns:
        Requests 'logger' object (mainly for unit testing)
    """
    http.client.HTTPConnection.debuglevel = 1
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.propagate = True
    requests_log.setLevel(logging.DEBUG)
    return requests_log


def exception_logger(func):
    """Explicitly log Exceptions then Raise them.

    Logging Exceptions and Tracebacks while inside of a thread is broken in the
    Tornado futures package for Python 2.7. It swallows most of the traceback
    and only gives you the raw exception object. This little helper method
    allows us to throw a log entry with the full traceback before raising the
    exception.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            log.debug('Exception caught in %s(%s, %s): %s' %
                      (func, args, kwargs, e), exc_info=1)
            raise
    return wrapper


def retry(excs, retries=3, delay=0.25):
    """Coroutine-compatible Retry Decorator.

    This decorator provides a simple retry mechanism that looks for a
    particular set of exceptions and retries async tasks in the event that
    those exceptions were caught.

    Example usage:
        >>> @gen.coroutine
        ... @retry(excs=(Exception), retries=3)
        ... def login(self):
        ...     raise gen.Return()

    Args:
        excs: A single (or tuple) exception type to catch.
        retries: The number of times to try the operation in total.
        delay: Time (in seconds) to wait between retries
    """
    def _retry_on_exc(f):
        def wrapper(*args, **kwargs):
            i = 1
            while True:
                try:
                    # Don't log the first time..
                    if i > 1:
                        log.debug('Try (%s/%s) of %s(%s, %s)' %
                                  (i, retries, f, args, kwargs))
                    ret = yield gen.coroutine(f)(*args, **kwargs)
                    log.debug('Result: %s' % ret)
                    raise gen.Return(ret)
                except excs as e:
                    log.error('Exception raised on try %s: %s' % (i, e))

                    if i >= retries:
                        log.debug('Raising exception: %s' % e)
                        raise e

                    i += 1
                    log.debug('Retrying in %s...' % delay)
                    yield tornado_sleep(delay)
                log.debug('Retrying..')
        return wrapper
    return _retry_on_exc


@gen.coroutine
def tornado_sleep(seconds=1.0):
    """Async method equivalent to sleeping.

    Args:
        seconds: Float seconds. Default 1.0
    """
    yield gen.sleep(seconds)


def populate_with_tokens(string, tokens, left_wrapper='%', right_wrapper='%',
                         strict=True, escape_sequence='\\',
                         remove_escape_sequence=True):
    """Insert token variables into the string.

    Will match any token wrapped in '%'s and replace it with the value of that
    token.

    Args:
        string: string to modify.
        tokens: dictionary of key:value pairs to inject into the string.
        left_wrapper: the character to use as the START of a token
        right_wrapper: the character to use as the END of a token
        strict: (bool) whether or not to make sure all tokens were replaced
        escape_sequence: character string to use as the escape sequence for
        left and right wrappers
        remove_escape_sequence: (bool) whether or not to remove the escape
        sequence if it found. For example \\%FOO\\% would turn into %FOO%.
    Example:
        export ME=biz

        string='foo %ME% %bar%'
        populate_with_tokens(string, os.environ)  # 'foo biz %bar%'
    """

    # First things first, swap out all instances of %<str>% with any matching
    # token variables found. If no items are in the hash (none, empty hash,
    # etc), then skip this.
    allowed_types = (str, str, bool, int, float)
    if tokens:
        for k, v in tokens.items():

            if type(v) not in allowed_types:
                log.warning('Token %s=%s is not in allowed types: %s' % (
                    k, v, allowed_types))
                continue

            string = string.replace(
                ('%s%s%s' % (left_wrapper, k, right_wrapper)), str(v))

    tokens_with_default = re.finditer(
        r'{0}(([\w]+)[|]([^{1}]+)){1}'.format(left_wrapper, right_wrapper),
        string)
    for match, key, default in (m.groups() for m in tokens_with_default):
        value = tokens.get(key, default)
        string = string.replace(
            '%s%s%s' % (left_wrapper, match, right_wrapper), str(value))

    # Slashes need to be escaped properly because they are a
    # part of the regex syntax.
    escape_sequence = escape_sequence.replace('\\', '\\\\')
    escape_pattern = r'({0}{1})([\w]+)({0}{2})'.format(
        escape_sequence,
        left_wrapper,
        right_wrapper)

    # If we are strict, we check if we missed anything. If we did, raise an
    # exception.
    if strict:
        missed_tokens = list(set(re.findall(r'%s[\w]+%s' %
                                 (left_wrapper, right_wrapper), string)))

        # Remove the escaped tokens from the missing tokens
        escape_findings = re.finditer(escape_pattern, string)

        escaped_tokens = [m.groups()[1] for m in escape_findings]
        missed_tokens = list(set(missed_tokens) - set(escaped_tokens))

        if missed_tokens:
            raise LookupError(
                'Found un-matched tokens in JSON string: %s' % missed_tokens)

    # Find text that's between the wrappers and escape sequence and
    # replace with just the wrappers and text.
    if remove_escape_sequence:
        string = re.sub(
            escape_pattern,
            r'{0}\2{1}'.format(left_wrapper, right_wrapper),
            string)
    return string


def convert_script_to_dict(script_file, tokens):
    """Converts a JSON file to a config dict.

    Reads in a JSON file, swaps out any environment variables that
    have been used inside the JSON, and then returns a dictionary.

    Args:
        script_file: Path to the JSON/YAML file to import, or file instance.
        tokens: dictionary to pass to populate_with_tokens.

    Returns:
        <Dictonary of Config Data>

    Raises:
        kingpin.exceptions.InvalidScript
    """

    filename = ''
    try:
        if isinstance(script_file, IOBase):
            filename = script_file.name
            instance = script_file
        else:
            filename = script_file
            instance = io.open(script_file)
    except IOError as e:
        raise exceptions.InvalidScript('Error reading script %s: %s' %
                                       (script_file, e))

    log.debug('Reading %s' % filename)
    raw = instance.read()
    parsed = populate_with_tokens(raw, tokens)

    # If the file ends with .json, use demjson to read it. If it ends with
    # .yml/.yaml, use PyYAML. If neither, error.
    suffix = filename.split('.')[-1].strip().lower()

    try:
        if suffix == 'json':
            decoded = demjson.decode(parsed)
        elif suffix in ('yml', 'yaml'):
            decoded = yaml.safe_load(parsed)
            if decoded is None:
                raise exceptions.InvalidScript(
                    'Invalid YAML in `%s`' % filename)
        else:
            raise exceptions.InvalidScriptName(
                'Invalid file extension: %s' % suffix)
    except demjson.JSONError as e:
        # demjson exceptions have `pretty_description()` method with
        # much more useful info.
        raise exceptions.InvalidScript('JSON in `%s` has an error: %s' % (
            filename, e.pretty_description()))
    return decoded


def order_dict(obj):
    """Re-orders a dict into a predictable pattern.

    Used so that you can compare two dicts with the same values, but that were
    created in different orders.

    Stolen from:
      http://stackoverflow.com/questions/25851183/how-to-compare-two-json-
      objects-with-the-same-elements-in-a-different-order-equa

    args:
        obj: Object to order

    returns:
        obj: A sorted version of the object
    """
    if isinstance(obj, dict):
        return sorted((k, order_dict(v)) for k, v in list(obj.items()))
    if isinstance(obj, list):
        return sorted((order_dict(x) for x in obj), key=str)
    else:
        return obj


def create_repeating_log(logger, message, handle=None, **kwargs):
    """Create a repeating log message.

    This function sets up tornado to repeatedly log a message in a way that
    does not need to be `yield`-ed.

    Example::

       >>> yield do_tornado_stuff(1)
       >>> log_handle = create_repeating_log('Computing...')
       >>> yield do_slow_computation_with_insufficient_logging()
       >>> clear_repeating_log(log_handle)

    This is similar to javascript's setInterval() and clearInterval().

    Args:
        message: String to pass to log.info()
        kwargs: values accepted by datetime.timedelta namely seconds, and
        milliseconds.

    Must be cleared via clear_repeating_log()
    Only handles one interval per actor.
    """

    class OpaqueHandle(object):

        """Tornado async io handler."""

        def __init__(self):
            self.timeout_id = None

    if not handle:
        handle = OpaqueHandle()

    def log_and_queue():
        logger(message)
        create_repeating_log(logger, message, handle, **kwargs)

    deadline = datetime.timedelta(**kwargs)
    # Here we only queue the call, we don't want to wait on it!
    timeout_id = ioloop.IOLoop.current().add_timeout(deadline, log_and_queue)
    handle.timeout_id = timeout_id

    return handle


def clear_repeating_log(handle):
    """Stops the timeout function from being called."""
    ioloop.IOLoop.current().remove_timeout(handle.timeout_id)


def diff_dicts(dict1, dict2):
    """Compares two dicts and returns the difference as a string,
    if there is any.

    Sorts two dicts (including sorting of the lists!!) and then diffs them.
    This will ignore string types ('unicode' vs 'string').

    args:
        dict1: First dict
        dict2: Second dict

    returns:
        A diff string if there's any difference, otherwise None.
    """
    dict1 = order_dict(dict1)
    dict2 = order_dict(dict2)

    if dict1 == dict2:
        return

    dict1 = pprint.pformat(dict1).splitlines()
    dict2 = pprint.pformat(dict2).splitlines()

    # Remove unicode identifiers.
    dict1 = [line.replace('u\'', '\'') for line in dict1]
    dict2 = [line.replace('u\'', '\'') for line in dict2]

    return '\n'.join(difflib.unified_diff(dict1, dict2, n=2))



"""Tests for the actors.spotinst package"""

from nose.plugins.attrib import attr
import uuid
import os

from tornado import testing

from kingpin.actors import spotinst

UUID = uuid.uuid4().hex


__author__ = 'Matt Wise <matt@nextdoor.com>'


class IntegrationSpotinstElastiGroup(testing.AsyncTestCase):

    """Integration tests against the Spotinst API.

    These tests actually hit the Spotinst API and test that the code
    works, as well as validate that the API credentials are working properly.

    Require environment variables:
      SPOTINST_TOKEN
      SECGRP
      SUBNET
      ZONE
      AMI
    """

    integration = True
    region = 'us-east-1'
    group_name = 'kingpin-%s' % UUID
    config = 'examples/test/spotinst.elastigroup/unittest.json'

    @attr('spotinst', 'integration', 'dry')
    @testing.gen_test(timeout=60)
    def integration_01a_create_elastigroup_dry(self):
        actor = spotinst.ElastiGroup(
            dry=True,
            init_tokens=os.environ,
            options={
                'name': self.group_name,
                'state': 'present',
                'config': self.config,
                'wait_on_create': True,
            })
        res = yield actor.execute()
        self.assertEqual(res, None)

    @attr('spotinst', 'integration')
    @testing.gen_test(timeout=300)
    def integration_01b_create_elastigroup(self):
        actor = spotinst.ElastiGroup(
            dry=False,
            init_tokens=os.environ,
            options={
                'name': self.group_name,
                'state': 'present',
                'config': self.config,
                'wait_on_create': True,
            })
        res = yield actor.execute()
        self.assertEqual(res, None)

    @attr('spotinst', 'integration', 'dry')
    @testing.gen_test(timeout=60)
    def integration_02a_update_elastigroup_dry(self):
        actor = spotinst.ElastiGroup(
            dry=True,
            init_tokens=os.environ,
            options={
                'name': self.group_name,
                'state': 'present',
                'config': self.config,
                'wait_on_create': True,
                'roll_on_change': True,
                'roll_batch_size': 99,
                'roll_grace_period': 60,
            })

        # Patch the group description to trigger an update
        actor._config['group']['description'] = UUID
        res = yield actor.execute()
        self.assertEqual(res, None)

    @attr('spotinst', 'integration')
    @testing.gen_test(timeout=1800)
    def integration_02b_update_elastigroup(self):
        actor = spotinst.ElastiGroup(
            dry=False,
            init_tokens=os.environ,
            options={
                'name': self.group_name,
                'state': 'present',
                'config': self.config,
                'roll_on_change': True,
                'wait_on_roll': True,
                'roll_batch_size': 99,
                'roll_grace_period': 60,
            })

        # Patch the group description to trigger an update
        actor._config['group']['description'] = UUID
        res = yield actor.execute()
        self.assertEqual(res, None)

    @attr('spotinst', 'integration', 'dry')
    @testing.gen_test(timeout=60)
    def integration_09a_delete_elastigroup(self):
        actor = spotinst.ElastiGroup(
            dry=False,
            init_tokens=os.environ,
            options={
                'name': self.group_name,
                'state': 'absent',
                'config': self.config,
            })
        res = yield actor.execute()
        self.assertEqual(res, None)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.alerts`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import logging

from tornado import gen
import requests

from kingpin.actors import exceptions
from kingpin.actors.utils import dry
from kingpin.actors.rightscale import base
from kingpin.constants import SchemaCompareBase
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


class InvalidInputs(exceptions.InvalidOptions):

    """Raised when supplied inputs are invalid for a ServerArray."""


class AlertSpecNotFound(exceptions.RecoverableActorFailure):

    """Raised when an Alert Spec could not be found"""


class CreateFailed(exceptions.RecoverableActorFailure):

    """Raised when an Alert Spec could not be created"""


class AlertsBaseActor(base.RightScaleBaseActor):

    """Abstract Alerts Actor that provides some utility methods."""

    @gen.coroutine
    def _find_alert_spec(self, name, subject_href):
        """Search for an AlertSpec by-name and return the resource.

        Note: A non-exact resource match is used below so that we return all of
        the Alert Specs that are matched by name. This method returns the
        resources in a list.

        Args:
            name: RightScale AlertSpec Name
            subject_href: The HREF of the subject this AlertSpec is assigned
              to.

        Return:
            [<rightcale.Resource objects>]
        """
        log.debug('Searching for AlertSpec matching: %s' % name)
        found_spec = yield self._client.find_by_name_and_keys(
            self._client._client.alert_specs, name, exact=False,
            subject_href=subject_href)

        if not found_spec:
            log.debug('AlertSpec matching "%s" could not be found.' % name)
            return

        log.debug('Got AlertSpec: %s' % found_spec)
        raise gen.Return(found_spec)


class Create(AlertsBaseActor):

    """Create a RightScale Alert Spec

    Options match the documentation in RightScale:
    http://reference.rightscale.com/api1.5/resources/ResourceAlertSpecs.html#create

    **Options**

    :array:
      The name of the Server or ServerArray to create the AlertSpec on.

    :strict_array:
      Whether or not to fail if the Server/ServerArray does not exist.
      (default: False)

    :condition:
      The condition (operator) in the condition sentence.
      (`>, >=, <, <=, ==, !=`)

    :description:
      The description of the AlertSpec.
      (*optional*)

    :duration:
      The duration in minutes of the condition sentence.
      (`^\d+$`)

    :escalation_name:
      Escalate to the named alert escalation when the alert is triggered.
      (*optional*)

    :file:
      The RRD path/file_name of the condition sentence.

    :name:
      The name of the AlertSpec.

    :threshold:
      The threshold of the condition sentence.

    :variable:
      The RRD variable of the condition sentence

    :vote_tag:
      Should correspond to a vote tag on a ServerArray if vote to grow or
      shrink.

    :vote_type:
      Vote to grow or shrink a ServerArray when the alert is triggered. Must
      either escalate or vote.
      (`grow` or `shrink`)

    **Examples**

    Create a high network activity alert on my-array:

    .. code-block:: json

       { "desc": "Create high network rx alert",
         "actor": "rightscale.alerts.Create",
         "options": {
           "array": "my-array",
           "strict_array": true,
           "condition": ">",
           "description": "Alert if amount of network data received is high",
           "duration": 180,
           "escalation_name": "Email Engineering",
           "file": "interface/if_octets-eth0",
           "name": "high network rx activity",
           "threshold": "50000000",
           "variable": "rx"
         }
       }

    **Dry Mode**

    In Dry mode this actor *does* validate that the ``array`` array exists.
    If it does not, a `kingpin.actors.rightscale.api.ServerArrayException` is
    thrown. Once that has been validated, the dry mode execution simply logs
    the Alert Spec that it would have created.

    Example *dry* output::

        TODO: Fill this in
    """

    all_options = {
        'array': (str, REQUIRED, 'Name of the ServerArray act on.'),
        'strict_array': (bool, False,
                         ('Whether or not to fail if the  Server/ServerArray ',
                          'does not exist.')),
        'condition': (str, REQUIRED,
                      'The condition (operator) in the condition sentence.'),
        'description': (str, None, 'The description of the AlertSpec.'),
        'duration': ((int, str), REQUIRED,
                     'The duration in minutes of the condition sentence.'),
        'escalation_name': (str, None,
                            ('Escalate to the named alert escalation when the',
                             'alert is triggered. Must either escalate or',
                             'vote.')),
        'file': (str, REQUIRED,
                 'The RRD path/file_name of the condition sentence.'),
        'name': (str, REQUIRED, 'The name of the AlertSpec.'),
        'threshold': (str, REQUIRED,
                      'The threshold of the condition sentence.'),
        'variable': (str, REQUIRED,
                     'The RRD variable of the condition sentence.'),
        'vote_tag': (str, None,
                     ('Should correspond to a vote tag on a ServerArray if ',
                      'vote to grow or shrink.')),
        'vote_type': (str, None,
                      ('Vote to grow or shrink a ServerArray when the alert ',
                       'is triggered. Must either escalate or vote.'))
    }

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""
        super(Create, self).__init__(*args, **kwargs)
        # By default, we're strict on our array array validation
        self._array_raise_on = 'notfound'
        self._array_allow_mock = False

        if not self.option('strict_array'):
            self._array_raise_on = None
            self._array_allow_mock = True

        if self.option('vote_type') not in ('grow', 'shrink', None):
            raise exceptions.InvalidOptions(
                'vote_type must be either: grow, shrink, None')

    @gen.coroutine
    def _execute(self):
        # Find the array we're adding an alert spec to. Specifically, we need
        # the servers HREF.
        array = yield self._find_server_arrays(
            self.option('array'),
            raise_on=self._array_raise_on,
            allow_mock=self._array_allow_mock)
        self.log.info('Found %s (%s)' % (array.soul['name'], array.href))

        # Add all of the required parameters to a dictionary
        params = {
            'condition': self.option('condition'),
            'description': self.option('description'),
            'duration': int(self.option('duration')),
            'file': self.option('file'),
            'name': self.option('name'),
            'subject_href': array.href,
            'threshold': self.option('threshold'),
            'variable': self.option('variable'),
        }

        # Generate the RightScale parameters that we need to pass in when
        # creating the alert. The optional parameters should not be passed in
        # if their option value came in as None.
        _optional_params = [
            'description', 'escalation_name', 'vote_tag', 'vote_type'
        ]
        for optional in _optional_params:
            if self.option(optional):
                params[optional] = self.option(optional)

        params = self._generate_rightscale_params('alert_spec', params)
        self.log.debug('Generated params: %s' % params)

        if self._dry:
            # In dry run mode, just log out what we would have done.
            self.log.info('Would have created the alert spec \"%s\" on %s' %
                          (self.option('name'), array.soul['name']))
            raise gen.Return()

        # We're really doin this. If we get a known exception back, handle
        # it. Otherwise, raise it.
        try:
            yield self._client.create_resource(
                self._client._client.alert_specs, params)
            self.log.info('Alert spec has been created')
        except requests.exceptions.HTTPError as e:
            if e.response.status_code in (422, 400):
                msg = ('Invalid parameters supplied to Alert Spec "%s": %s'
                       % (self.option('name'), params))
                raise exceptions.RecoverableActorFailure(msg)
            raise


class Destroy(AlertsBaseActor):

    """Destroy existing RightScale Alert Specs

    This actor searches RightScale for any Alert Specs that match the ``name``
    and ``array`` that you supplied, then deletes all of them. RightScale lets
    you have multiple alert specs with the same name, so if this actor finds
    multiple specs, it will delete them all.

    **Options**

    :array:
      The name of the Server or ServerArray to delete the AlertSpec from.

    :name:
      The name of the AlertSpec.

    **Examples**

    Destroy a high network activity alert on my-array:

    .. code-block:: json

       { "desc": "Destroy high network rx alert",
         "actor": "rightscale.alerts.Destroy",
         "options": {
           "array": "my-array",
           "name": "high network rx activity",
         }
       }

    **Dry Mode**

    In Dry mode this actor *does* validate that the ``array`` array exists,
    and that the AlertSpec exists on that array so that it can be deleted. A
    RecoverableActorFailure error is thrown if it does not exist.

    Example *dry* output::

        14:31:49   INFO      Rehearsing... Break a leg!
        14:31:49   INFO      [DRY: Kingpin] Preparing actors from delete.json
        14:31:53   INFO      [DRY: Destroy high network rx alert] Found
          my-array (/api/server_arrays/329142003) to delete alert spec from
        14:31:54   INFO      [DRY: Destroy high network rx alert] Would have
          deleted the alert spec "high network rx activity" on my-array
    """

    all_options = {
        'array': (str, REQUIRED, 'Name of the ServerArray act on.'),
        'name': (str, REQUIRED, 'The name of the AlertSpec.')
    }

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""
        super(Destroy, self).__init__(*args, **kwargs)
        # By default, we're strict on our array validation
        self._array_raise_on = 'notfound'
        self._array_allow_mock = False

    @gen.coroutine
    def _execute(self):
        # Find the array we're adding an alert spec to. Specifically, we need
        # the servers HREF.
        array = yield self._find_server_arrays(
            self.option('array'),
            raise_on=self._array_raise_on,
            allow_mock=self._array_allow_mock)

        self.log.info('Found %s (%s) to delete alert spec from' %
                      (array.soul['name'], array.href))

        # Find the AlertSpec on this server, if it exists.
        alerts = yield self._find_alert_spec(
            self.option('name'), array.href)

        # If we can't find the AlertSpec specific to the subjet array that was
        # supplied, raise an exception and bail.
        if not alerts:
            raise AlertSpecNotFound(
                '"%s" could not be found on %s' %
                (self.option('name'), array.soul['name']))

        # We'll store our 'delete spec' futures in here
        deletes = []

        for spec in alerts:
            log.debug('Found Alert Spec %s' % spec.soul)

            if self._dry:
                # In dry run mode, just log out what we would have done.
                self.log.info('Would have deleted alert \"%s\" (%s) on %s' %
                              (spec.soul['name'],
                               spec.href,
                               array.soul['name']))
            else:
                # We're really doin this!
                self.log.info('Deleting alert \"%s\" (%s) on %s' %
                              (spec.soul['name'],
                               spec.href,
                               array.soul['name']))
                deletes.append(self._client.destroy_resource(spec))

        # Wait for the deletes to finish
        if deletes:
            yield deletes


class AlertSpecSchema(SchemaCompareBase):

    """Provides JSON-Schema based verification of the supplied AlertSpec

    The majority of the schema mirrors the RightScale API:
      http://reference.rightscale.com/api1.5/resources/ResourceAlertSpecs.html#create
    """

    SCHEMA = {
        'type': ['object'],
        'required': [
            'condition', 'duration', 'file', 'name', 'threshold', 'variable'
        ],
        'properties': {
            'condition': {
                'type': 'string',
                'enum': ['>', '>=', '<', '<=', '==', '!=']
            },
            'description': {'type': 'string'},
            'duration': {'type': 'integer'},
            'escalation_name': {'type': 'string'},
            'file': {'type': 'string'},
            'name': {'type': 'string'},
            'threshold': {'type': 'string'},
            'variable': {'type': 'string'},
            'vote_tag': {'type': 'string'},
            'vote_type': {'enum': ['grow', 'shrink']}
        }
    }


class AlertSpecsSchema(SchemaCompareBase):

    """Provides JSON-Schema verification that the supplied input was a list
    of AlertSpecSchemas."""

    SCHEMA = {
        'type': ['array', 'null'],
        'uniqueItems': True,
        'items': {
            'anyOf': [AlertSpecSchema.SCHEMA]
        }
    }


class AlertSpecBase(base.EnsurableRightScaleBaseActor):

    """Extremely simple AlertSpec creation actor.

    This actor isn't really meant to be instantiated on its own -- it provides
    the base functionality though for creating, deleting and updating an
    AlertSpec on a given RightScale resource. The resource can be either a
    Server Array, Server Template, Instance or Deployment.

    **Options**

    :href:
      The RightScale HREF for the resource you wish to apply the Alert Spec to.

    :state:
      (str) Either `present` or `absent`

    :spec:
      A dictionary that conforms to the
      :py:mod:`~kingpin.actors.rightscale.alerts.AlertSpecSchema`.

    **Examples**

    .. code-block:: json

       { "actor": "rightscale.alerts.AlertSpecBase",
         "options": {
           "href": "/api/server_arrays/abcd1234",
           "spec": {
                "name": "Instance Stranded",
                "description": "Alert if an instance enders a stranded",
                "file": "RS/server-failure",
                "variable": "state",
                "condition": "==",
                "threshold": "stranded",
                "duration": 2,
                "escalation_name": "critical"
           }
         }
       }

    """

    all_options = {
        'href': (str, None, 'RightScale Resource HREF to act on'),
        'spec': (AlertSpecSchema, None,
                 'The actual Alert Spec definition itself.')
    }
    unmanaged_options = ['href']
    desc = "AlertSpec: {spec[name]}"

    def __init__(self, *args, **kwargs):
        super(AlertSpecBase, self).__init__(*args, **kwargs)
        self.changed = False

    @gen.coroutine
    def _precache(self):
        name = self.option('spec').get('name')
        href = self.option('href')

        # Generate a fully populated set of parameters by including the href
        # that was supplied to us.
        desired_spec = dict(self.option('spec'))
        desired_spec['subject_href'] = href
        self.desired_params = self._generate_rightscale_params(
            'alert_spec', desired_spec)

        # Search for the existing spec. Even though we do an 'exact' search
        # here, this is kind of misleading. We are searching on multiple keys
        # (name, subject_href) and the find_by_name_and_keys() code isn't smart
        # enough to return a true perfect match on all the keys supplied.
        # Therefore, this MAY return a list with multiple results.
        log.debug('Searching for AlertSpec matching: %s' % name)
        self.existing_spec = yield self._client.find_by_name_and_keys(
            self._client._client.alert_specs, exact=True,
            name=name, subject_href=href)

        # Handle the multiple-result scenario. Since exact=True above, we
        # should never get a list back with a single item, so we'll only focus
        # on the issue of getting back multiple items. Search through the list
        # of items for an _exact_ match on the name.
        if (isinstance(self.existing_spec, list) and
                len(self.existing_spec) > 1):
            self.existing_spec = [
                s for s in self.existing_spec if
                s.soul['name'] == self.option('spec')['name']][0]

        if not self.existing_spec:
            log.debug('AlertSpec matching "%s" could not be found.' % name)
            self.existing_spec = None
        else:
            log.debug('Got AlertSpec: %s' % self.existing_spec)

    @gen.coroutine
    def _get_state(self):
        if self.existing_spec:
            raise gen.Return('present')
        raise gen.Return('absent')

    @gen.coroutine
    def _set_state(self):
        if self.option('state') == 'absent':
            yield self._delete_spec()
        else:
            yield self._create_spec()

        self.changed = True

    @gen.coroutine
    def _get_spec(self):
        if self.existing_spec:
            raise gen.Return(
                self._strip_returned_spec_resource(self.existing_spec))

    @gen.coroutine
    def _set_spec(self):
        yield self._update_spec()
        self.changed = True

    @gen.coroutine
    @dry('Would have created the AlertSpec')
    def _create_spec(self):
        try:
            self.existing_spec = yield self._client.create_resource(
                self._client._client.alert_specs, self.desired_params)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code in (422, 400):
                msg = ('Invalid parameters supplied to Alert Spec "%s": %s'
                       % (self.option('href'), self.desired_params))
                raise exceptions.RecoverableActorFailure(msg)
            raise

        self.log.info('Alert spec has been created')

    @gen.coroutine
    @dry('Would have updated the AlertSpec')
    def _update_spec(self):
        try:
            self.existing_spec = yield self._client.update(
                self.existing_spec, self.desired_params)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code in (422, 400):
                msg = ('Invalid parameters supplied to Alert Spec "%s": %s'
                       % (self.existing_spec.soul['name'],
                          self.desired_params))
                raise exceptions.RecoverableActorFailure(msg)
            raise

        self.log.info('Alert spec has been updated')

    @gen.coroutine
    @dry('Would have deleted the AlertSpec')
    def _delete_spec(self):
        yield self._client.destroy_resource(self.existing_spec)
        self.existing_spec = None

        self.log.info('Alert spec has been destroyed')

    def _strip_returned_spec_resource(self, spec):
        """Converts an AlertSpec resource into a comparable dict.

        Walks through the AlertSpecSchema and creates a new dictionary object
        with only the values from the AlertSpec that were supplied. This
        creates a dict we can compare against the supplied data by the user.

        args:
            spec: A RightScale AlertSpec resource

        returns:
            <dictionary that matches the AlertSpecSchema>
        """
        new = {}
        desired_keys = list(AlertSpecSchema.SCHEMA['properties'].keys())
        for key in desired_keys:
            if key in spec.soul:
                new[key] = spec.soul[key]
        return new


class AlertSpecsBase(base.EnsurableRightScaleBaseActor):

    """Extremely simple AlertSpec management actor.

    This actor isn't really meant to be instantiated on its own -- it provides
    the base functionality though for creating, deleting and updating an
    AlertSpec on a given RightScale resource. The subtle difference here is
    that this actor manages the entire list of AlertSpecs on a given resource,
    rather than just a single spec.

    **Options**

    :href:
      The RightScale HREF for the resource you wish to apply the Alert Spec to.

    :state:
      (str) Either `present` or `absent`

    :spec:
      A dictionary that conforms to the
      :py:mod:`~kingpin.actors.rightscale.alerts.AlertSpecSchema`.

    **Examples**

    .. code-block:: json

       { "actor": "rightscale.alerts.AlertSpecsBase",
         "options": {
           "href": "/api/server_arrays/abcd1234",
           "specs": [
                { "name": "Instance Stranded",
                  "description": "Alert if an instance enders a stranded",
                  "file": "RS/server-failure",
                  "variable": "state",
                  "condition": "==",
                  "threshold": "stranded",
                  "duration": 2,
                  "escalation_name": "critical"
                }
            ]
         }
       }

    """

    all_options = {
        'href': (str, None, 'RightScale Resource HREF to act on'),
        'specs': (AlertSpecsSchema, None, 'A list of AlertSpec dicts'),
    }
    unmanaged_options = ['href']
    desc = "AlertSpecs: {href}"

    def __init__(self, *args, **kwargs):
        super(AlertSpecsBase, self).__init__(*args, **kwargs)
        self.changed = False

        self.alert_actors = []
        for spec in self.option('specs'):
            a = AlertSpecBase(
                desc=('AlertSpecs: %s "%s"' % (
                      self.option('href'), spec['name'])),
                options={
                    'href': self.option('href'),
                    'state': self.option('state'),
                    'spec': spec
                },
                dry=self._dry)

            # Replace the new child actors RightScale client with our own --
            # this way we only have to log in once rather than
            # once-per-alert-spec.
            a._client = self._client
            self.alert_actors.append(a)

    @gen.coroutine
    def _precache(self):
        # Get a list of all of the AlertSpecs associated with the resource.
        all_resource_specs = yield self._client.find_by_name_and_keys(
            self._client._client.alert_specs, exact=True,
            subject_href=self.option('href'))
        if not isinstance(all_resource_specs, list):
            all_resource_specs = [all_resource_specs]

        # Now quickly compare this list to the list of desired specs. For each
        # one thats found that doesn't match, create a new AlertSpecBase actor
        # that will purge this AlertSpec.
        desired_spec_names = [s['name'] for s in self.option('specs')]

        for spec in all_resource_specs:
            if spec.soul['name'] in desired_spec_names:
                continue

            # Create a AlertSpec actor where state is absent
            self.alert_actors.append(
                AlertSpecBase(
                    desc=('AlertSpecs: %s "%s"' % (
                          self.option('href'), spec.soul['name'])),
                    options={
                        'href': self.option('href'),
                        'state': 'absent',
                        'spec': {
                            'name': spec.soul['name'],

                            # The rest of these options don't matter, but need
                            # to be here in order to pass the Schema
                            # verification.
                            'condition': '==',
                            'duration': 0,
                            'file': 'bogus',
                            'threshold': 'NaN',
                            'variable': 'state'
                        }
                    },
                    dry=self._dry))

        # Now create all of the AlertSpec actors that we want to use to manage
        # the defined alert specs.
        tasks = []
        for actor in self.alert_actors:
            # Do all the heavy lifting by throwing these _precache() calls onto
            # a task list and then yield the whole list at once. This makes the
            # rest of the get/set comparison calls super fast.
            tasks.append(actor._precache())
        yield tasks

    @gen.coroutine
    def _get_state(self):
        raise gen.Return()

    @gen.coroutine
    def _compare_state(self):
        equals = True

        # For every actor we've created, make sure the state matches the
        # desired state. If they aren't, set our equals to False and let
        # set_state() get called.
        for actor in self.alert_actors:
            # Note, this is super fast because the AlertSpecBase._precache()
            # actor did all of the actual API calls ahead of time. This returns
            # instantly.
            is_equal = yield actor._compare_state()
            if not is_equal:
                equals = False

        raise gen.Return(equals)

    @gen.coroutine
    def _set_state(self):
        tasks = []
        for actor in self.alert_actors:
            is_equal = yield actor._compare_state()
            if not is_equal:
                tasks.append(actor._set_state())
                self.changed = True
        yield tasks

    @gen.coroutine
    def _get_specs(self):
        specs = []
        for actor in self.alert_actors:
            # Note, this is super fast because the AlertSpecBase._precache()
            # actor did all of the actual API calls ahead of time. This returns
            # instantly.
            specs.append((yield actor._get_spec()))
        raise gen.Return(specs)

    @gen.coroutine
    def _set_specs(self):
        tasks = []
        for actor in self.alert_actors:
            # Note, this is super fast because the AlertSpecBase._precache()
            # actor did all of the actual API calls ahead of time. This returns
            # instantly.
            equals = yield actor._compare_spec()
            if not equals:
                # Note, we don't call _set_spec() here intentionally because
                # that doesn't have the logic to skip setting the spec if we
                # want it absent.
                tasks.append(actor._execute())
                self.changed = True
        yield tasks



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.rightscript`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import logging

from tornado import gen

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.rightscale import base
from kingpin.actors.utils import dry
from kingpin.constants import REQUIRED, STATE

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


class RightScript(base.EnsurableRightScaleBaseActor):

    """Manages the state of a RightScale Script

    Options match the documentation in RightScale:
    http://reference.rightscale.com/api1.5/resources/ResourceRightScripts.html

    **RightScript Inputs**
    RightScripts have a concept of "inputs" ... pieces of data that are
    supplied to the script on-startup to help configure your instance. By
    *default*, RightScale parses the script you attach and automatically
    exposes inputs for you:

    http://docs.rightscale.com/cm/rs101/understanding_inputs.html

    RightScale has recently released an automatic script parser that looks for
    YAML-formatted comments in your script and reads the metadata from the
    comments. This is much more powerful and allows you to be explicit about
    the inputs required for your script to function.

    http://docs.rightscale.com/cm/dashboard/design/rightscripts/rightscripts_metadata_comments.html

    **Options**

    :name:
      (str) The name of the RightScript.

    :description:
      (str) Optional description of the RightScript.

    :packages:
      (list, str) A list of packages that need to be installed in order to run
      the script.

    :source:
      (str) A file name with the contents of the script you want to upload.
      Script should contain the *RightScale Metadata Comments* (see above)
      which will automatically handle configuring the inputs for your script.

    :commit:
      (str) Optional comment used to commit the revision if Kingpin makes any
      changes at all.

    **Examples**

    Create a high network activity alert on my-array:

    .. code-block:: json

        { "actor": "rightscale.alerts.Create",
          "options": {
            "name": "Set Hostname",
            "ensure": "present",
            "commit": "yep",
            "description": "Set the hostname to something usable",
            "packages": "hostname sed curl",
            "source": "./set_hostname.sh"
          }
        }
    """

    all_options = {
        'name': (str, REQUIRED, 'Name of the RightScript to manage'),
        'state': (STATE, 'present',
                  'The condition (operator) in the condition sentence.'),
        'commit': (str, None, 'Commit the RightScript revision on-change.'),
        'description': (str, None, 'The description of the RightScript.'),
        'packages': (str, None, 'Space separaged list of packages to install'),
        'source': (str, REQUIRED, 'File containing the script contents.'),
    }

    unmanaged_options = ['commit', 'name']

    desc = 'RightScript: {name}'

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""
        super(RightScript, self).__init__(*args, **kwargs)
        self.changed = False

        # The rightscale API allows you to push an invalid list of packages
        # (multiple spaces, newlines, etc). We need to sanitize the list
        # quickly first.
        if self.option('packages'):
            self._options['packages'] = ' '.join(
                self.option('packages').split())

        self._desired_source = self._read_source()
        self._desired_params = self._generate_rightscale_params(
            prefix='right_script',
            params={
                'description': self.option('description'),
                'name': self.option('name'),
                'packages': self.option('packages'),

                # Passing in right_script[source] is required at creation time
                # .. but we can't pass in large scripts this way. Instead, we
                # pass in bogus text here, and rely on set_source() to properly
                # update the source with a special API endpoint. This value is
                # stripped out of any update_params() calls, so its only used
                # at creation time.
                'source': 'bogus'
            })

    def _read_source(self):
        """Reads in a RightScript source file.

        Reads in the file contents, and swaps in any tokens that may have
        been left in the script.

        args:
            source: Path to the script

        returns:
            <string contents of the script>
        """
        source = self.option('source')

        try:
            fh = open(source)
            raw = fh.read()
        except IOError as e:
            raise exceptions.InvalidOptions('Error reading script %s: %s' %
                                            (source, e))

        try:
            parsed = utils.populate_with_tokens(raw, self._init_tokens)
        except LookupError as e:
            raise exceptions.InvalidOptions('Error parsing tokens in %s: %s' %
                                            (source, e))

        return parsed

    @gen.coroutine
    def _precache(self):
        # First go off and find our script object
        name = self.option('name')
        log.debug('Searching for RightScript matching: %s' % name)
        found = yield self._client.find_by_name_and_keys(
            self._client._client.right_scripts, exact=True, name=name)

        if not found:
            log.debug('RightScript matching "%s" could not be found.' % name)
            self.script = None
            self.source = None
            raise gen.Return()

        # It was found, so save a reference to the object
        if not isinstance(found, list):
            found = [found]

        log.debug('Got RightScript: %s' % found[0])
        self.script = found[0]

        # Next, get the source of the script
        self.source = yield self._client.make_generic_request(
            self.script.source.path)

    @gen.coroutine
    def _set_state(self):
        """Creates a RightScript.

        args:
            name: The name of the script to create
        """
        if self._dry:
            self.log.warning('Would have set RightScript state: %s' %
                             self.option('state'))
            self.script = None
            self.changed = True
            raise gen.Return()

        if self.option('state') == 'absent':
            if not self.script:
                raise gen.Return()

            self.log.info('Destroying RightScript')
            yield self._client.destroy_resource(self.script)
            self.script = None
            self.changed = True
            raise gen.Return()

        self.log.info('Creating RightScript')
        self.script = yield self._client.create_resource(
            self._client._client.right_scripts, self._desired_params)
        self.changed = True

    @gen.coroutine
    def _get_state(self):
        if self.script is None:
            raise gen.Return('absent')

        raise gen.Return('present')

    @gen.coroutine
    @dry('Would have updated the RightScript parameters')
    def _update_params(self):
        self.log.info('Updating RightScript parameters...')
        params = [t for t in self._desired_params if
                  t[0] != 'right_script[source]']
        self.script = yield self._client.update(self.script, params)
        self.changed = True

    @gen.coroutine
    @dry('Would have updated the RightScale Source')
    def _set_source(self):
        self.log.info('Updating RightScript source...')
        self.script = yield self._client.update(
            self.script,
            self._desired_source,
            sub_resource='source')
        self.changed = True

    @gen.coroutine
    def _get_source(self):
        raise gen.Return(self.source)

    @gen.coroutine
    def _compare_source(self):
        existing = yield self._get_source()
        equals = (self._desired_source == existing)
        raise gen.Return(equals)

    @gen.coroutine
    def _set_description(self):
        self.log.warning('Descriptions do not match')
        yield self._update_params()

    @gen.coroutine
    def _get_description(self):
        if self.script is None or 'description' not in self.script.soul:
            raise gen.Return()

        raise gen.Return(self.script.soul['description'])

    @gen.coroutine
    def _set_packages(self):
        self.log.warning('Packages do not match')
        yield self._update_params()

    @gen.coroutine
    def _get_packages(self):
        if self.script is None or 'packages' not in self.script.soul:
            raise gen.Return()

        raise gen.Return(self.script.soul['packages'])

    @gen.coroutine
    @dry('Would have committed HEAD to a revision')
    def _commit(self):
        self.log.info('Committing a new revision')

        ret = yield self._client.commit_resource(
            res=self.script, res_type=self._client._client.right_scripts,
            params={'right_script[commit_message]': self.option('commit')})

        self.log.info('Committed revision %s' % ret.soul['revision'])

    @gen.coroutine
    def _execute(self):
        yield super(RightScript, self)._execute()

        if self.option('state') == 'absent':
            raise gen.Return()

        # Finally, if we're committing and a change was made, commit!
        if self.changed and self.option('commit'):
            yield self._commit()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.settings`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Common settings used by many of the `kingpin.actors.aws` modules.
"""


import os

import boto

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'

# By default, this means that Boto will make HTTP calls at instantiation time
# to determine whether or not credentials are available from the metadata
# service.
#
# During tests, we mock these out to blank strings to prevent these calls.
AWS_ACCESS_KEY_ID = os.getenv('AWS_ACCESS_KEY_ID', None)
AWS_SECRET_ACCESS_KEY = os.getenv('AWS_SECRET_ACCESS_KEY', None)

SQS_RETRY_DELAY = 30

ECS_RETRY_ATTEMPTS = 3
ECS_RETRY_DELAY = 5


# Common Settings for the retrying.retry() decorator
#
# Use like this: @retrying.retry(**settings.RETRYING_SETTINGS)
#
def is_retriable_exception(exception):
    """Return true if this AWS exception is transient and should be retried.

    Example:
        >>> @retry(retry_on_exception=is_retriable_exception)
    """
    retry_codes = (
        'Throttling',
        'Rate exceeded',
        'reached max retries'
    )

    # Only handle Boto exceptions
    if not isinstance(exception, boto.exception.BotoServerError):
        return False

    # Boto exceptions should have a code attribute
    error_code = exception.error_code or ''
    return any([c in error_code for c in retry_codes])


RETRYING_SETTINGS = {
    # Verify if we need to retry with the is_retriable_exception
    # method described above.
    'retry_on_exception': is_retriable_exception,

    # Wait up to 10 times
    'stop_max_attempt_number': 10,

    # Add 250ms of random jitter to every retry
    'wait_jitter_max': 250,

    # Add 250ms of sleep to every retry
    'wait_fixed': 250,

    # Now add between 250-2000ms to every retry
    'wait_random_min': 250,
    'wait_random_max': 2000,

    # Finally, add in an exponential backoff timer with a 10s limit
    'wait_exponential_multiplier': 100,
    'wait_exponential_max': 10000
}
